import torch
import torch.nn as nn
import math

class AttentionModule(nn.Module):
  '''A class used for implementing the general transformer attention mechanism.'''
  def __init__(self,
               is_decoder=False,
               num_positional_encoding_buckets=32,
               positional_encoding_max_distance=128,
               d_model=768,
               num_heads=12,
               dropout_rate=0.1,
               has_positional_encoding=False):
        """
        Constructs an AttentionModule instance with specific hyperparameters.

        :param is_decoder: Indicates if we are using a decoder.
        :type is_decoder: bool, optional

        :param num_positional_encoding_buckets: Number of positional encoding buckets.
        :type num_positional_encoding_buckets: int, optional

        :param positional_encoding_max_distance: Max distance for positional encoding.
        :type positional_encoding_max_distance: int, optional

        :param d_model: Indicates the model embeddings dimension.
        :type d_model: int, optional

        :param num_heads: States the number of attention heads.
        :type num_heads: int, optional

        :param dropout_rate: Dropout rate.
        :type dropout_rate: float, optional

        :param has_positional_encoding: If positional encoding is applied.
        :type has_positional_encoding: bool, optional

        :return: None
        """
        super().__init__()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.is_decoder = is_decoder
        self.has_positional_encoding = has_positional_encoding
        self.num_positional_encoding_buckets = num_positional_encoding_buckets
        self.positional_encoding_max_distance = positional_encoding_max_distance
        self.d_model = d_model
        self.per_head_dim = d_model // num_heads
        self.n_heads = num_heads
        self.dropout = dropout_rate

        self.q = nn.Linear(self.d_model, self.d_model, bias=False)
        self.k = nn.Linear(self.d_model, self.d_model, bias=False)
        self.v = nn.Linear(self.d_model, self.d_model, bias=False)
        self.o = nn.Linear(self.d_model, self.d_model, bias=False)

        if self.has_positional_encoding:
            self.relative_attention_bias = nn.Embedding(self.num_positional_encoding_buckets, self.n_heads)

  def relative_positional_encoding(self, relative_position, bidirectional=True, num_buckets=32, max_distance=128):
      """
      Provides the buckets given the relative positions.

      :param relative_position: Tensor of relative positions.
      :type relative_position: Tensor

      :param bidirectional: If the attention is bidirectional, is false in  the decoder as the token can attend only to the tokens behid it.
      :type bidirectional: bool, optional

      :param num_buckets: Number of buckets for positional encoding.
      :type num_buckets: int, optional

      :param max_distance: Maximum distance for positional encoding.
      :type max_distance: int, optional

      :return: Relative buckets.
      :rtype: Tensor
      """
      relative_buckets = 0
      if bidirectional: # self attention in encoder
          num_buckets //= 2
          relative_buckets += (relative_position > 0).to(torch.long) * num_buckets
          relative_position = torch.abs(relative_position)
      else: # self attention in decoder (you can't attend to what is yet to come) for efficient utilization of buckets
          relative_position = -torch.min(relative_position, torch.zeros_like(relative_position).to(self.device))
      
      # now all is in the positive positions realm (mapping to buckets is much straightforward)
      max_exact = num_buckets // 2
      is_small = relative_position < max_exact

      relative_position_if_large = max_exact + (
          torch.log(relative_position.float() / max_exact)
          / math.log(max_distance / max_exact)
          * (num_buckets - max_exact)
      ).to(torch.long)
      relative_position_if_large = torch.min(
          relative_position_if_large, torch.full_like(relative_position_if_large, num_buckets - 1).to(self.device)
      )

      relative_buckets += torch.where(is_small, relative_position, relative_position_if_large)
      return relative_buckets

  def compute_bias(self, query_length, key_length):
      """
      Computes the the relative positional bias between the queries and the keys.

      :param query_length: Length of the query sequance.
      :type query_length: int

      :param key_length: Length of the key sequance.
      :type key_length: int

      :return: Positional embeddings.
      :rtype: Tensor
      """
      self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
      context_position = torch.arange(query_length, dtype=torch.long, device=self.device)[:, None]
      memory_position = torch.arange(key_length, dtype=torch.long, device=self.device)[None, :]
      relative_position = memory_position - context_position
      
      relative_position_bucket = self.relative_positional_encoding(
          relative_position,
          bidirectional=(not self.is_decoder),
          num_buckets=self.num_positional_encoding_buckets,
          max_distance=self.positional_encoding_max_distance,
      )
      positional_embeddings = self.relative_attention_bias(relative_position_bucket) # mapping bucketes to their corresponding embeddings
      positional_embeddings = positional_embeddings.permute([2, 0, 1]).unsqueeze(0)
      return positional_embeddings

  def forward(
      self,
      hidden_states,
      mask=None,
      key_value_states=None,
      position_bias=None):
      """
      The forward pass of the attention layer.

      :param hidden_states: Tensor of the Query.
      :type hidden_states: Tensor

      :param mask: Mask to be applied on values.
      :type mask: Tensor, optional

      :param key_value_states: Tensor of the Key and Value, the default is the same as hidden_states.
      :type key_value_states: Tensor, optional

      :param position_bias: Positional bias to be added.
      :type position_bias: Tensor, optional

      :return: Returns the attention output and positional bias.
      :rtype: Tuple[Tensor, Tensor]
      """
      batch_size, seq_length = hidden_states.shape[:2]

      real_seq_length = seq_length

      key_length = real_seq_length if key_value_states is None else key_value_states.shape[1]

      def shape(states):
          return states.view(batch_size, -1, self.n_heads, self.per_head_dim).transpose(1, 2)

      def unshape(states):
          return states.transpose(1, 2).contiguous().view(batch_size, -1, self.d_model)

      query_states = shape(self.q(hidden_states))

      if key_value_states is not None:
        key_states= shape(self.k(key_value_states))
        value_states = shape(self.v(key_value_states))
      else:
        key_states= shape(self.k(hidden_states))
        value_states = shape(self.v(hidden_states))

      scores = torch.matmul( # getting the weightes over the whole batch over all the heads and all at once
          query_states, key_states.transpose(3, 2)
      )

      if position_bias is None:
          if not self.has_positional_encoding:
              position_bias = torch.zeros(
                  (1, self.n_heads, real_seq_length, key_length), device=scores.device, dtype=scores.dtype
              ).to(self.device)
          else:
              position_bias = self.compute_bias(real_seq_length, key_length)

          if mask is not None:
              position_bias = position_bias + mask # mask here is not 0 and 1 but -inf and 0

      position_bias_masked = position_bias

      scores += position_bias_masked
      attn_weights = nn.functional.softmax(scores.float(), dim=-1).type_as(
          scores
      )
      attn_weights = nn.functional.dropout(
          attn_weights, p=self.dropout, training=self.training
      )

      attn_output = unshape(torch.matmul(attn_weights, value_states))
      attn_output = self.o(attn_output)

      return attn_output, position_bias

class NewGELUActivation(nn.Module):
    '''Simple interface of the Gaussian Error Linear Units (GELU) activation function'''
    def forward(self, input):
        return 0.5 * input * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (input + 0.044715 * torch.pow(input, 3.0))))


class DenseGatedActDenseModule(nn.Module):
  '''A class used to implement a dense, gated activation function.'''
  def __init__(self,
               d_model=768,
               d_ff=2048,
               dropout_rate=0.1):
        '''
        Initializes the DenseGatedActDense Module class with the given parameters which is a gated dense layer followed a dense layer.

        :param d_model: Input dimension to the module (and also the model embedding dimension).
        :type d_model: int, optional

        :param d_ff: Hidden layer dimension.
        :type d_ff: int, optional

        :param dropout_rate: Dropout rate.
        :type dropout_rate: float, optional

        :return: None
        '''
        super().__init__()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.wi_0 = nn.Linear(d_model, d_ff, bias=False)
        self.wi_1 = nn.Linear(d_model, d_ff, bias=False)
        self.wo = nn.Linear(d_ff, d_model, bias=False)
        self.dropout = nn.Dropout(dropout_rate)
        self.act = NewGELUActivation()

  def forward(self, hidden_states):
      """
      Performs the forward pass of the dense, gated activation function.

      :param hidden_states: Input tensor to the forward method.
      :type hidden_states: Tensor

      :return: Output tensor after applying dense, gated activation function.
      :rtype: Tensor
      """
      hidden_gelu = self.act(self.wi_0(hidden_states))
      hidden_linear = self.wi_1(hidden_states)
      hidden_states = hidden_gelu * hidden_linear
      hidden_states = self.dropout(hidden_states)
      hidden_states = self.wo(hidden_states)
      return hidden_states

class LayerNormModule(nn.Module):
  """
  A class used to apply the Layer Normalization operation.
  """
  def __init__(self, layer_size = 768, eps=1e-6):
    '''
    Initializes the Layer Normalization Module class with the given parameters.

    :param layer_size: Dimensions of the layers to be normalized.
    :type layer_size: int, optional

    :param eps: A term added to improve numerical stability.
    :type eps: float, optional

    :return: None
    '''
    super().__init__()
    self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    self.weight = nn.Parameter(torch.ones(layer_size).to(self.device))
    self.epsilon = eps
  def forward(self, hidden_states):
    mean_square = hidden_states.to(torch.float32).pow(2).mean(-1, keepdim=True)
    hidden_states = hidden_states * torch.rsqrt(mean_square + self.epsilon)
    return self.weight * hidden_states

class FFModule(nn.Module):
  '''A class used to execute a Feed-Forward Neural Network module.'''
  def __init__(self,
                dropout_rate=0.1):
        '''
        Initializes the FFModule class with the given parameters (the feed-forward block in the Transformer).

        :param hidden_states: Input tensor to the forward method.
        :type hidden_states: Tensor

        :return: Output tensor after applying the FFN.
        :rtype: Tensor
        '''
        super().__init__()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.DenseReluDense = DenseGatedActDenseModule()
        self.layer_norm = LayerNormModule()
        self.dropout = nn.Dropout(dropout_rate)

  def forward(self, hidden_states):
      """
      Perform the forward pass of the Feed-forward Neural Network module.

      :param hidden_states: Input tensor to the Feed Forward Network (FFN).
      :type hidden_states: Tensor

      :return: Output tensor after applying the FFN.
      :rtype: Tensor
      """
      forwarded_states = self.layer_norm(hidden_states)
      forwarded_states = self.DenseReluDense(forwarded_states)
      hidden_states = hidden_states + self.dropout(forwarded_states)
      return hidden_states

class SelfAttentionModule(nn.Module):
  '''A class used to implement a self-attention mechanism.'''
  def __init__(self,
               is_decoder,
               dropout_rate=0.1,
               has_positional_encoding=False):
    """
    Initializes the SelfAttentionModule class with the given parameters.

    :param is_decoder: Indicates if we are using a decoder (the decoder and encoder has differant ways to hadle the positional encoding).
    :type is_decoder: bool

    :param dropout_rate: Dropout rate.
    :type dropout_rate: float, optional

    :param has_positional_encoding: If positional encoding is applied.
    :type has_positional_encoding: bool, optional

    :return: None
    """
    super().__init__()
    self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    self.SelfAttention = AttentionModule(is_decoder, has_positional_encoding=has_positional_encoding)
    self.layer_norm = LayerNormModule()
    self.dropout = nn.Dropout(dropout_rate)

  def forward(self, hidden_states, attention_mask=None, position_bias=None):
      """
      Applies the self-attention to the hidden states.

      :param hidden_states: Tensor of the Query, Key and Value (all have the same inpus as this is self attention).
      :type hidden_states: Tensor

      :param attention_mask: Attention mask for the self-attention mechanism.
      :type attention_mask: Tensor, optional

      :param position_bias: Position bias for self-attention.
      :type position_bias: Tensor, optional

      :return: Returns the hidden states and position bias.
      :rtype: Tuple[Tensor, Tensor]
      """
      normed_hidden_states = self.layer_norm(hidden_states)
      attention_output = self.SelfAttention(
          normed_hidden_states,
          mask=attention_mask,
          position_bias=position_bias
      )
      hidden_states = hidden_states + self.dropout(attention_output[0])
      position_bias = attention_output[1]
      return hidden_states, position_bias

class EncoderBlock(nn.Module):
  '''A class used for the encoder block of the transformer model.'''
  def __init__(self, has_positional_encoding=False):
    """
    Initializes the EncoderBlock class with the given parameters.

    :param has_positional_encoding: If positional encoding is applied.
    :type has_positional_encoding: bool, optional

    :return: None
    """
    super().__init__()
    self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    self.layer = nn.ModuleList()
    self.layer.append(SelfAttentionModule(is_decoder=False, has_positional_encoding=has_positional_encoding))
    self.layer.append(FFModule())

  def forward(self, hidden_states, attention_mask=None, position_bias=None):
    """
    Encoder block forward pass.

    :param hidden_states: Input tensor to the Encoder block.
    :type hidden_states: Tensor

    :param attention_mask: Attention mask for the self-attention mechanism.
    :type attention_mask: Tensor, optional

    :param position_bias: Position bias for self-attention.
    :type position_bias: Tensor, optional

    :return: Returns the hidden states and position bias.
    :rtype: Tuple[Tensor, Tensor]
    """
    self_attention_outputs = self.layer[0](hidden_states, attention_mask=attention_mask, position_bias=position_bias)
    hidden_states = self_attention_outputs[0]
    position_bias = self_attention_outputs[1]
    hidden_states = self.layer[-1](hidden_states)
    return hidden_states, position_bias

class EncoderModule(nn.Module):
  '''A class used for the encoder of the transformer model.'''
  def __init__(self, embed_tokens, num_layers= 12, dropout_rate=0.1):
    """
    Initializes the EncoderModule class with the given parameters.

    :param embed_tokens: The embeddings of the input tokens.
    :type embed_tokens: nn.Embedding
    
    :param num_layers: The number of encoder layers in the model.
    :type num_layers: int, optional
    
    :param dropout_rate: The dropout rate.
    :type dropout_rate: float, optional

    :return: None
    """
    super().__init__()
    self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    self.embed_tokens = embed_tokens
    self.block = nn.ModuleList(
            [EncoderBlock(has_positional_encoding=bool(i == 0)) for i in range(num_layers)]
        )
    self.final_layer_norm = LayerNormModule()
    self.dropout = nn.Dropout(dropout_rate)

  def forward(self, input_ids=None, attention_mask=None):
    """
    Performs the forward pass of the encoder module.

    :param input_ids: The indices of the input sequence tokens.
    :type input_ids: Tensor, optional
    
    :param attention_mask: The binary mask indicating the positions where the input sequence is padded (1 for not padded, 0 for padded).
    :type attention_mask: Tensor, optional

    :return: The encoded hidden states.
    :rtype: Tensor
    """
    input_shape = input_ids.size()
    input_ids = input_ids.view(-1, input_shape[-1])

    inputs_embeds = self.embed_tokens(input_ids)
    
    if attention_mask is None:
      attention_mask = torch.ones(input_shape).to(self.device)
    final_attention_mask = attention_mask.to(dtype=torch.float32)
    final_attention_mask = (1.0 - final_attention_mask) * torch.finfo(torch.float32).min

    hidden_states = self.dropout(inputs_embeds)
    position_bias = None
    for block in self.block:
      layer_outputs = block(
        hidden_states,
        attention_mask = final_attention_mask,
        position_bias = position_bias
        )
      hidden_states = layer_outputs[0]
      position_bias = layer_outputs[1]

    hidden_states = self.final_layer_norm(hidden_states)
    hidden_states = self.dropout(hidden_states)
    return hidden_states

class CrossAttentionModule(nn.Module):
  '''A class used for the cross-attention module of the transformer model.'''
  def __init__(self, dropout_rate=0.1):
    """
    Initializes the CrossAttentionModule class with the given parameters.

    :param dropout_rate: Dropout rate.
    :type dropout_rate: float, optional

    :return: None
    """
    super().__init__()
    self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    self.EncDecAttention = AttentionModule(is_decoder=True, has_positional_encoding=False)
    self.layer_norm = LayerNormModule()
    self.dropout = nn.Dropout(dropout_rate)

  def forward(self, hidden_states, key_value_states, encoder_attention_mask=None):
      """
      Applies cross-attention where the query comes from the hidden states and the key and value come from key_value_states.

      :param hidden_states: Input tensor to be used for the query.
      :type hidden_states: Tensor

      :param key_value_states: Input tensor to be used for the key and value.
      :type key_value_states: Tensor
      
      :param encoder_attention_mask: Attention mask for the cross-attention mechanism.
      :type encoder_attention_mask: Tensor, optional

      :return: Returns the hidden states and position bias.
      :rtype: Tuple[Tensor, Tensor]
      """
      normed_hidden_states = self.layer_norm(hidden_states)
      attention_output = self.EncDecAttention(
          normed_hidden_states,
          mask=encoder_attention_mask,
          key_value_states=key_value_states
      )
      hidden_states = hidden_states + self.dropout(attention_output[0])
      position_bias = attention_output[1]
      return hidden_states, position_bias

class DecoderBlock(nn.Module):
  '''A class used for the decoder block of the transformer model.'''
  def __init__(self, has_positional_encoding=False):
        """
        Initializes the DecoderModule class with the given parameters.

        :param has_positional_encoding: If positional encoding is applied.
        :type has_positional_encoding: bool, optional

        :return: None
        """
        super().__init__()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.layer = nn.ModuleList()
        self.layer.append(SelfAttentionModule(is_decoder=True, has_positional_encoding=has_positional_encoding))
        self.layer.append(CrossAttentionModule())
        self.layer.append(FFModule())

  def forward(
      self,
      hidden_states,
      attention_mask=None,
      position_bias=None,
      encoder_hidden_states=None,
      encoder_attention_mask=None
  ):
      """
      Performs the forward pass of the decoder block.

      :param hidden_states: The hidden states from the previous decoder block (or the input embeddings if this is the first decoder block).
      :type hidden_states: Tensor

      :param attention_mask: The binary mask of the decoder sequance, indicating the positions where the input sequence is padded (1 for not padded, 0 for padded).
      :type attention_mask: Tensor, optional
      
      :param position_bias: The positional bias for self-attention mechanism.
      :type position_bias: Tensor, optional

      :param encoder_hidden_states: The output hidden states from the encoder module.
      :type encoder_hidden_states: Tensor, optional
      
      :param encoder_attention_mask: The binary mask of the encoder sequance, indicating where the input has been padded.
      :type encoder_attention_mask: Tensor, optional

      :return: The hidden states and position bias after the forward pass of the decoder block.
      :rtype: Tuple[Tensor, Tensor]
      """
      self_attention_outputs = self.layer[0](
          hidden_states,
          attention_mask=attention_mask,
          position_bias=position_bias
      )
      hidden_states = self_attention_outputs[0]
      position_bias = self_attention_outputs[1]

      cross_attention_outputs = self.layer[1](
          hidden_states,
          key_value_states=encoder_hidden_states,
          encoder_attention_mask=encoder_attention_mask
      )
      hidden_states = cross_attention_outputs[0]

      hidden_states = self.layer[-1](hidden_states)
      return hidden_states, position_bias

class DecoderModule(nn.Module):
  '''A class used to implement the decoder part of the transformer.'''
  def __init__(self, embed_tokens, num_layers= 12, dropout_rate=0.1):
    """
    Initializes the DecoderModule class with the given parameters.

    :param embed_tokens: The embeddings of the decoder input sequance tokens.
    :type embed_tokens: nn.Embedding

    :param num_layers: The number of decoder layers in the model.
    :type num_layers: int, optional
    
    :param dropout_rate: The dropout rate.
    :type dropout_rate: float, optional

    :return: None
    """
    super().__init__()
    self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    self.embed_tokens = embed_tokens
    self.block = nn.ModuleList(
        [DecoderBlock(has_positional_encoding=bool(i == 0)) for i in range(num_layers)]
    )
    self.final_layer_norm = LayerNormModule()
    self.dropout = nn.Dropout(dropout_rate)

  def forward(
              self,
              input_ids=None,
              attention_mask=None,
              encoder_hidden_states=None,
              encoder_attention_mask=None
  ):
    """
    Performs the forward pass of the decoder.

    :param input_ids: The indices of the input sequence tokens in the vocabulary.
    :type input_ids: Tensor, optional
    
    :param attention_mask: The binary mask of the decoder sequance, indicating the positions where the input sequence is padded (1 for not padded, 0 for padded).
    :type attention_mask: Tensor, optional

    :param encoder_hidden_states: the output of the encoder module.
    :type encoder_hidden_states: Tensor, optional
    
    :param encoder_attention_mask: The binary mask of the encoder sequance, indicating the positions where the input sequence is padded (1 for not padded, 0 for padded).
    :type encoder_attention_mask: Tensor, optional

    :return: Decoded output hidden states.
    :rtype: Tensor
    """
    input_shape = input_ids.size()
    input_ids = input_ids.view(-1, input_shape[-1])
    inputs_embeds = self.embed_tokens(input_ids)

    if attention_mask is None:
      attention_mask = torch.ones(input_shape).to(self.device)
    seq_length = input_shape[1]
    lower_triangular_mask = torch.tril(torch.ones((seq_length, seq_length))).view(1, 1, seq_length, seq_length).to(self.device)
    final_attention_mask = lower_triangular_mask * attention_mask
    final_attention_mask = (1.0 - final_attention_mask) * torch.finfo(torch.float32).min
    
    if encoder_attention_mask is None:
      encoder_attention_mask = torch.ones((encoder_hidden_states.size(0), encoder_hidden_states.size(1))).to(self.device)
    encoder_attention_mask = encoder_attention_mask.to(dtype=torch.float32)
    encoder_attention_mask = (1.0 - encoder_attention_mask) * torch.finfo(torch.float32).min

    hidden_states = self.dropout(inputs_embeds)
    position_bias = None
    for block in self.block:
      layer_outputs = block(
          hidden_states,
          attention_mask=final_attention_mask,
          position_bias=position_bias,
          encoder_hidden_states=encoder_hidden_states,
          encoder_attention_mask=encoder_attention_mask
      )

      hidden_states = layer_outputs[0]
      position_bias = layer_outputs[1]
    hidden_states = self.final_layer_norm(hidden_states)
    hidden_states = self.dropout(hidden_states)
    return hidden_states

class T5Model(nn.Module):
    '''A class to represent the T5 transformer model.'''
    def __init__(self, vocab_size=32128, d_model=768):
        """
        Initializes the T5 Model with the given parameters.

        :param vocab_size: The size of the vocabulary.
        :type vocab_size: int, optional

        :param d_model: The dimensionality of the input embedding.
        :type d_model: int, optional

        :return: None
        """
        super().__init__()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.shared = nn.Embedding(vocab_size, d_model)
        self.encoder = EncoderModule(self.shared)
        self.decoder = DecoderModule(self.shared)
        self.lm_head = nn.Linear(d_model, vocab_size, bias=False)
        self.softmax = nn.Softmax(dim=-1)
        self.loss_fn = nn.CrossEntropyLoss(ignore_index=-100)
        
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        decoder_input_ids=None,
        decoder_attention_mask=None):
        """
        Performs the forward pass of the T5 transformer model.

        :param input_ids: The IDs of the input tokens.
        :type input_ids: Tensor, optional
        
        :param attention_mask: The binary mask of the encoder sequance, indicating the positions where the input sequence is padded (1 for not padded, 0 for padded).
        :type attention_mask: Tensor, optional

        :param decoder_input_ids: The ids of the decoder input tokens.
        :type decoder_input_ids: Tensor, optional

        :param decoder_attention_mask: The binary mask of the decoder sequance, indicating the positions where the input sequence is padded (1 for not padded, 0 for padded).
        :type decoder_attention_mask: Tensor, optional

        :return: The token probabilities of the output sequence.
        :rtype: Tensor
        """
        # encoder
        encoder_outputs = self.encoder(
            input_ids=input_ids,
            attention_mask=attention_mask
        )
        # decoder
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_outputs,
            encoder_attention_mask=attention_mask,
        )
        # lm_head
        lm_logits = self.lm_head(decoder_outputs)
        output = self.softmax(lm_logits)
        loss = self.loss_fn(lm_logits.view(-1, lm_logits.size(-1)), decoder_input_ids.view(-1))
        return output, loss
    
    def generate(
      self,
      input_ids=None,
      attention_mask=None,
      max_length=10,
      temperature=1.0):
      """
      Generates output sequence given input_ids and attention_mask.

      :param input_ids: The IDs of the input tokens.
      :type input_ids: Tensor, optional
      
      :param attention_mask: The binary mask of the encoder sequance, indicating the positions where the input sequence is padded (1 for not padded, 0 for padded).
      :type attention_mask: Tensor, optional

      :param max_length: The maximum length of the sequence to be generated.
      :type max_length: int, optional

      :param temperature: The temperature of the softmax function, the higher its value the flatter the probability distribution of the next token will be.
      :type temperature: float, optional

      :return: The IDs of the generated tokens.
      :rtype: Tensor
      """
      # put model in evaluation mode (stop dropout and backwrad propagation calculation)
      self.eval()
      encoder_outputs = self.encoder(
          input_ids=input_ids,
          attention_mask=attention_mask
      )
      # autoregressive generation
      generated_ids = [torch.tensor([0]).to(self.device)]
      cur_ids = torch.zeros((input_ids.size(0), 1)).long().to(self.device)
      # genrate until max_length or eos_token
      for _ in range(max_length):
        decoder_outputs = self.decoder(
            input_ids=cur_ids,
            encoder_hidden_states=encoder_outputs
        )
        lm_logits = self.lm_head(decoder_outputs)
        next_token_logits = lm_logits[:, -1, :] / temperature
        next_token_id = torch.multinomial(torch.softmax(next_token_logits, dim=-1), num_samples=1)[0]
        generated_ids.append(next_token_id)
        cur_ids = torch.cat([cur_ids, next_token_id.unsqueeze(-1)], dim=-1)
        if next_token_id == 1:
          break
      return torch.cat(generated_ids, dim=-1)



import re
import string
import numpy as np


class RawDataInstance():
  """
  Represents a raw data instance.

  :param dial_idx: Dialogue index.
  :type dial_idx: str
  :param turn_idx: Turn index.
  :type turn_idx: int
  :param user_utter: User utterance.
  :type user_utter: str
  :param sys_utter: System utterance.
  :type sys_utter: str
  :param history: Dialogue history.
  :type history: list[str]
  :param turn_slots: Slots for the current turn.
  :type turn_slots: dict[str, str]
  :param inform_mem: Informed slots from previous turns.
  :type inform_mem: dict[str, list[str]]
  """
  def __init__(self,
               dial_idx,
               turn_idx,
               user_utter,
               sys_utter,
               history,
               turn_slots,
               inform_mem):
    self.dial_idx = dial_idx
    self.turn_idx = turn_idx
    self.user_utter = user_utter
    self.sys_utter = sys_utter
    self.history = history
    self.turn_slots = turn_slots
    self.inform_mem = inform_mem

  def __str__(self):
    """
    Return a string representation of the RawDataInstance object.

    :return: A string representation of the object.
    :rtype: str
    """
    string = ''
    string = string + '\ndial_idx: ' + str(self.dial_idx)
    string = string + '\nturn_idx: ' + str(self.turn_idx)
    string = string + '\nuser_utter: ' + str(self.user_utter)
    string = string + '\nsys_utter: ' + str(self.sys_utter)
    string = string + '\nhistory: ' + str(self.history)
    string = string + '\nturn_slots: ' + str(self.turn_slots)
    string = string + '\ninform_mem: ' + str(self.inform_mem)
    return string

class DataInstance():
  """
  Represents a processed data instance.

  :param ids: Input IDs.
  :type ids: list[int]
  :param mask: Attention mask.
  :type mask: list[int]
  :param token_type_ids: Token type IDs.
  :type token_type_ids: list[int]
  :param spans: Spans.
  :type spans: list[int]
  :param spans_start: Start positions of spans.
  :type spans_start: list[int]
  :param spans_end: End positions of spans.
  :type spans_end: list[int]
  :param padding_len: Padding length.
  :type padding_len: int
  :param input_tokens: Input tokens.
  :type input_tokens: str
  :param input: Input text.
  :type input: str
  :param opers: Slot operations.
  :type opers: list[int]
  :param target_values: Target slot values.
  :type target_values: list[str]
  :param last_state: Last dialogue state.
  :type last_state: dict[str, str]
  :param cur_state: Current dialogue state.
  :type cur_state: dict[str, str]
  :param refer: Referenced slots.
  :type refer: list[int]
  :param inform_aux_features: Informed auxiliary features.
  :type inform_aux_features: list[float]
  :param ds_aux_features: Filled slot auxiliary features.
  :type ds_aux_features: list[float]
  """
  def __init__(self,
               ids,
               mask,
               token_type_ids,
               spans,
               spans_start,
               spans_end,
               padding_len,
               input_tokens,
               input,
               opers,
               target_values,
               last_state,
               cur_state,
               refer,
               inform_aux_features,
               ds_aux_features):
    self.ids = ids
    self.mask = mask
    self.token_type_ids = token_type_ids
    self.spans = spans
    self.spans_start = spans_start
    self.spans_end = spans_end
    self.padding_len = padding_len
    self.input_tokens = input_tokens
    self.input = input
    self.opers = opers
    self.target_values = target_values
    self.last_state = last_state
    self.cur_state = cur_state
    self.refer = refer
    self.inform_aux_features = inform_aux_features
    self.ds_aux_features = ds_aux_features

  def __str__(self):
    """
    Return a string representation of the DataInstance object.

    :return: A string representation of the object.
    :rtype: str
    """
    string = ''
    string = string + '\nids: ' + str(self.ids)
    string = string + '\nmask: ' + str(self.mask)
    string = string + '\ntoken_type_ids: ' + str(self.token_type_ids)
    string = string + '\nspans: ' + str(self.spans)
    string = string + '\nspans_start: ' + str(self.spans_start)
    string = string + '\nspans_end: ' + str(self.spans_end)
    string = string + '\npadding_len: ' + str(self.padding_len)
    string = string + '\ninput_tokens: ' + str(self.input_tokens)
    string = string + '\ninput: ' + str(self.input)
    string = string + '\nopers: ' + str(self.opers)
    string = string + '\ntarget_values: ' + str(self.target_values)
    string = string + '\nlast_state: ' + str(self.last_state)
    string = string + '\ncur_state: ' + str(self.cur_state)
    string = string + '\nrefer: ' + str(self.refer)
    string = string + '\ninform_aux_features: ' + str(self.inform_aux_features)
    string = string + '\nds_aux_features: ' + str(self.ds_aux_features)

    return string

class AverageMeter():
    """
    Computes and stores the average and current value.
    """
    def __init__(self):
        self.reset()

    def reset(self):
        """
        Reset the average meter.
        """
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """
        Update the average meter with a new value.

        :param val: New value.
        :type val: float
        :param n: Number of instances the value represents.
        :type n: int
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def normalize(text, multiwoz):
  """
  Normalize the given text by converting it to lowercase and splitting it into tokens.

  :param text: Input text.
  :type text: str
  :return: Normalized tokens.
  :rtype: list[str]
  """
  text_lower = text.lower()
  if multiwoz == True:
    text_norm = normalize_text(text_lower) # for mutliwoz only
  else:
    text_norm = text_lower
  text_tok = [tok for tok in map(lambda x: re.sub(" ", "", x), re.split("(\W+)", text_norm)) if len(tok) > 0]
  return text_tok


def is_included(value, target):
  """
  Check if the target is included in the value.

  :param value: The value to check.
  :type value: str
  :param target: The target value to search for.
  :type target: str
  :return: True if the target is included in the value, False otherwise.
  :rtype: bool
  """
  included = False

  value = [item for item in map(str.strip, re.split("(\W+)", value)) if len(item) > 0]
  target = [item for item in map(str.strip, re.split("(\W+)", target)) if len(item) > 0]

  for i in range(len(value)):
    if value[i:i + len(target)] == target:
      included = True

  return included


def included_with_label_maps(value, target, label_maps):
  """
  Check if the value is included in the target or any of its variants based on the label maps.

  :param value: The value to check.
  :type value: str
  :param target: The target value to search for.
  :type target: str
  :param label_maps: Dictionary of label maps.
  :type label_maps: dict[str, list[str]]
  :return: True if the value is included in the target or any of its variants, False otherwise.
  :rtype: bool
  """
  included = False

  variants = [target]
  if target in label_maps:
    variants += label_maps[target]

  for variant in variants:
    if value == variant or is_included(value, variant) or is_included(variant, value):
      included = True

  return included


def match_with_label_maps(value, target, label_maps={}):
    """
    Check if the value matches the target or any of its variants based on the label maps.

    :param value: The value to check.
    :type value: str
    :param target: The target value to match against.
    :type target: str
    :param label_maps: Dictionary of label maps.
    :type label_maps: dict[str, list[str]]
    :return: True if the value matches the target or any of its variants, False otherwise.
    :rtype: bool
    """
    equal = False
    if value == target:
      equal = True
    elif target in label_maps:
      for variant in label_maps[target]:
        if value == variant:
          equal = True

    return equal


def create_span_output(output_start, output_end, padding_len, input_tokens):
  """
  Create the span output based on the output start and end positions.

  :param output_start: Output start positions.
  :type output_start: list[int]
  :param output_end: Output end positions.
  :type output_end: list[int]
  :param padding_len: Padding length.
  :type padding_len: int
  :param input_tokens: Input tokens.
  :type input_tokens: str
  :return: The created span output.
  :rtype: str
  """
  mask = [0] * (len(output_start) - padding_len)

  if padding_len > 0:
    idx_start = np.argmax(output_start[1:-padding_len]) + 1
    idx_end = np.argmax(output_end[1:-padding_len]) + 1
  else:
    idx_start = np.argmax(output_start[1:]) + 1
    idx_end = np.argmax(output_end[1:]) + 1

  for mj in range(idx_start, idx_end + 1):
    mask[mj] = 1

  output_tokens = [x for p, x in enumerate(input_tokens.split()) if mask[p] == 1]
  output_tokens = [x for x in output_tokens if x not in ('[CLS]', '[SEP]')]

  final_output = ''
  for ot in output_tokens:
    if ot.startswith('##'):
      final_output = final_output + ot[2:]
    elif len(ot) == 1 and ot in string.punctuation:
      final_output = final_output + ot
    elif len(final_output) > 0 and final_output[-1] in string.punctuation:
      final_output = final_output + ot
    else:
      final_output = final_output + " " + ot

  final_output = final_output.strip()

  return final_output


def mask_utterance(utter, inform_mem, multiwoz, replace_with='[UNK]'):
  """
  Mask the utterance by replacing the informed values in the inform memory.

  :param utter: The utterance to mask.
  :type utter: list[str]
  :param inform_mem: The inform memory containing slot-value pairs.
  :type inform_mem: dict[str, list[str]]
  :param replace_with: The replacement token.
  :type replace_with: str
  :return: The masked utterance.
  :rtype: list[str]
  """
  utter = normalize(utter, multiwoz)
  for slot, informed_values in inform_mem.items():
    for informed_value in informed_values:
      informed_tok = normalize(informed_value, multiwoz)
      for i in range(len(utter)):
        if utter[i:i + len(informed_tok)] == informed_tok:
          utter[i:i + len(informed_tok)] = [replace_with] * len(informed_tok)
  return utter


def normalize_time(text):
    """
    Normalize the time format in the given text (specific to MultiWoz dataset).

    :param text: The input text.
    :type text: str
    :return: The normalized text.
    :rtype: str
    """
    
    # This code is only related to MultiWoz Dataset

    text = re.sub("(\d{1})(a\.?m\.?|p\.?m\.?)", r"\1 \2", text) # am/pm without space
    text = re.sub("(^| )(\d{1,2}) (a\.?m\.?|p\.?m\.?)", r"\1\2:00 \3", text) # am/pm short to long form
    text = re.sub("(^| )(at|from|by|until|after) ?(\d{1,2}) ?(\d{2})([^0-9]|$)", r"\1\2 \3:\4\5", text) # Missing separator
    text = re.sub("(^| )(\d{2})[;.,](\d{2})", r"\1\2:\3", text) # Wrong separator
    text = re.sub("(^| )(at|from|by|until|after) ?(\d{1,2})([;., ]|$)", r"\1\2 \3:00\4", text) # normalize simple full hour time
    text = re.sub("(^| )(\d{1}:\d{2})", r"\g<1>0\2", text) # Add missing leading 0
    # Map 12 hour times to 24 hour times
    text = re.sub("(\d{2})(:\d{2}) ?p\.?m\.?", lambda x: str(int(x.groups()[0]) + 12 if int(x.groups()[0]) < 12 else int(x.groups()[0])) + x.groups()[1], text)
    text = re.sub("(^| )24:(\d{2})", r"\g<1>00:\2", text) # Correct times that use 24 as hour
    return text


def normalize_text(text):
    """
    Normalize the text (specific to MultiWoz dataset).

    :param text: The input text.
    :type text: str
    :return: The normalized text.
    :rtype: str
    """

    # This code is only related to MultiWoz Dataset

    text = normalize_time(text)
    text = re.sub("n't", " not", text)
    text = re.sub("(^| )zero(-| )star([s.,? ]|$)", r"\g<1>0 star\3", text)
    text = re.sub("(^| )one(-| )star([s.,? ]|$)", r"\g<1>1 star\3", text)
    text = re.sub("(^| )two(-| )star([s.,? ]|$)", r"\g<1>2 star\3", text)
    text = re.sub("(^| )three(-| )star([s.,? ]|$)", r"\g<1>3 star\3", text)
    text = re.sub("(^| )four(-| )star([s.,? ]|$)", r"\g<1>4 star\3", text)
    text = re.sub("(^| )five(-| )star([s.,? ]|$)", r"\g<1>5 star\3", text)
    text = re.sub("archaelogy", "archaeology", text) # Systematic typo
    text = re.sub("guesthouse", "guest house", text) # Normalization
    text = re.sub("(^| )b ?& ?b([.,? ]|$)", r"\1bed and breakfast\2", text) # Normalization
    text = re.sub("bed & breakfast", "bed and breakfast", text) # Normalization
    text = re.sub("\t", " ", text) # Error
    text = re.sub("\n", " ", text) # Error
    return text



from flask import Flask, render_template, request, jsonify
from pyngrok import ngrok


class chat_gui():
    chat_func = None
    bot_type = None
    # one of ['Whiz Bot', 'Basic Bot', 'Task Bot', 'Converse Bot', 'Voice Bot', 'Theorizer']
    def __init__(self, bot_type, chat_func, collab=False, auth_token=None):
        chat_gui.bot_type = bot_type
        chat_gui.chat_func = chat_func
        if collab:
            ngrok.set_auth_token("2EN6nBHTHxD9kKRZsRFpdn4nNj2_d81ejeYV3pm844WLQXoi")
            if auth_token:  ngrok.set_auth_token(auth_token)
            ngrok_tunnel = ngrok.connect(5000)
            print('Public URL:', ngrok_tunnel.public_url)
        app.run()

app = Flask(__name__)

@app.route("/")
def index():
    return render_template('chat.html', bot=chat_gui.bot_type)


@app.route("/get", methods=["GET", "POST"])
def chat():
    msg = request.form["msg"]
    input = msg
    return get_Chat_response(input)


def get_Chat_response(text):
    return chat_gui.chat_func(text)





from gensim.utils import tokenize
import numpy as np
from nltk.stem.porter import PorterStemmer
stemmer = PorterStemmer()


class BoW():
    '''
    An interface for transforming sentences into bag-of-words vectors.
    '''
    def __init__(self, binary=False):
        self.all_words = None
        self.binary = binary
        
    def transform_list(self, sentence_list, all_words):
        '''
        Given a list of tokenized sentences, return a table of BoW vectors (one for each sentence) in the form of a numpy array.
        '''
        self.all_words = all_words
        BoWs = np.zeros((len(sentence_list), len(all_words)), dtype=np.float64)
        for i, sentence in enumerate(sentence_list):
            sentence = list(tokenize(sentence, to_lower=True))
            sentence = [stemmer.stem(word.lower()) for word in sentence if word not in ['?', '!', '.', ',']]
            for word in sentence:
                word_index = all_words.index(word)
                BoWs[i, word_index] += 1 if not self.binary else 1
        return BoWs
    
    def transform(self, sentence):
        '''
        Given a sentence, return its BoW vector as a numpy array.
        :param: sentence: A string of words
        '''
        sentence = list(tokenize(sentence, to_lower=True))
        sentence = [stemmer.stem(word.lower()) for word in sentence if word not in ['?', '!', '.', ',']]
        BoW = np.zeros((1, len(self.all_words)), dtype=np.float64)
        for word in sentence:
            if word not in self.all_words:  continue
            word_index = self.all_words.index(word)
            BoW[0, word_index] += 1 if not self.binary else 1
        return BoW


import pandas as pd
import torch
from tokenizers import Tokenizer
from tokenizers.trainers import BpeTrainer
from tokenizers.pre_tokenizers import Whitespace
from tokenizers.models import BPE

class WhizBot_GRU_Preprocessor:
    '''An interface that provides the required preprocessing for the WhizBot_GRU bot'''
    def __init__(self, file_path):
        """
        Constructs a WhizBot_GRU_Preprocessor instance with the file path of the dataset.

        :param file_path: Path to the .json file if the dataset.
        :type file_path: str

        :returns: None
        """
        self.file_path = file_path
        self.text_col = "text"
        self.label_col = "label"
        self.data = pd.read_json(self.file_path)
        # make a dictionary of all the tag and responses
        self.responces = {}
        for i in range(len(self.data)):
            self.responces[self.data['tag'][i]] = self.data['responses'][i]
        # remove the responses column from the dataframe
        self.data = self.data.drop('responses', axis=1)
        # stretch the dataframe to have one row per pattern
        self.data = self.data.explode('patterns')
        # rename the columns tag->label and patterns->text
        self.data = self.data.rename(columns={'tag': 'label', 'patterns': 'text'})
        # remove all rows with empty text
        self.data = self.data.dropna()
        self.data = self.data.reset_index(drop=True)

    def process(self):
        """
        loads the data, cleans it, tokenizes it, pads it and removes outlier sequances.

        :returns: Processed data.
        :rtype: DataFrame
        """
        # text processing
        self.data['text'] = self.data['text'].apply(self.clean_string)
        # byte pair encoding tokenizer
        self.tokenizer = Tokenizer(BPE())
        self.tokenizer.pre_tokenizer = Whitespace()
        # padding token is set to 0
        trainer = BpeTrainer(vocab_size=2000, min_frequency=2, special_tokens=["<pad>"])
        self.data['text'].to_csv("text.txt", index=False, header=False)
        self.tokenizer.train(["text.txt"], trainer)
        # tokenize the data
        self.data['text'] = self.data['text'].apply(self.tokenize_string)
        # remove outliers (too long or too short sequences)
        median = self.data['text'].apply(lambda x: len(x)).median()
        iqr = self.data['text'].apply(lambda x: len(x)).quantile(0.75) - self.data['text'].apply(lambda x: len(x)).quantile(0.25)
        self.data = self.data[self.data['text'].apply(lambda x: len(x) >= median - 1.5 * iqr)]
        self.data = self.data[self.data['text'].apply(lambda x: len(x) <= median + 1.5 * iqr)]
        self.data = self.data.reset_index(drop=True)
        # pad the text sequences
        padding_id = self.tokenizer.token_to_id("<pad>")
        self.longest_sequence = self.data['text'].apply(lambda x: len(x)).max()
        self.data['text'] = self.data['text'].apply(lambda x: torch.cat((x, torch.zeros(self.longest_sequence - len(x)).fill_(padding_id).long())))
        # make a dictionary of all labels
        self.label_dict = {}
        for i in range(len(self.data)):
            if self.data['label'][i] not in self.label_dict:
                self.label_dict[self.data['label'][i]] = len(self.label_dict)
        # now replace all labels with their corresponding index
        self.data['label'] = self.data['label'].apply(lambda x: self.label_dict[x])
        # now convert to tensors
        self.data['label'] = self.data['label'].apply(lambda x: torch.tensor([x]))
        return self.data

    # process a single string
    def clean_string(self, string):
        """
        Cleans the given text string by removing punctuation, converting to lowercase and removing non-ascii characters.

        :param string: Provided string.
        :type string: str

        :returns: Cleaned string.
        :rtype: str
        """
        string = string.lower()
        string = string.replace('[^\w\s]','')
        string = string.encode("ascii", "ignore").decode()
        return string

    # tokenize a single string
    def tokenize_string(self, string):
        """
        Tokenizes the given text string.

        :param string: Provided string.
        :type string: str

        :returns: Tokenizens Id.
        :rtype: Tensor
        """
        tokens_vector = self.tokenizer.encode(string).ids
        tokens_vector = torch.tensor(tokens_vector)
        return tokens_vector

    # process a single string
    def process_string(self, string):
        """
        Cleans and tokenizes a given text string, the pads it to the longest sequence length (for batch processing).

        :param string: Provided string.
        :type string: str

        :returns: Processed padded tokens ids.
        :rtype: Tensor
        """
        string = self.clean_string(string)
        tokens_vector = self.tokenize_string(string)
        padded_tokens_vector = torch.cat((tokens_vector, torch.zeros(self.longest_sequence - len(tokens_vector)).long()))
        return padded_tokens_vector

    # pad a sequence (to make batch compatible)
    def pad_sequence(self, sequence):
        """
        Pads a given sequence to make it compatible with batch processing.

        :param sequence: Provided sequence.
        :type sequence: Tensor

        :returns: Padded sequence.
        :rtype: Tensor
        """
        return torch.nn.utils.rnn.pad_sequence(sequence.tolist(), batch_first=True, padding_value=self.tokenizer.token_to_id("[PAD]"))



"""
This Module has the run functions for TRIPPY that train and evaluate the model.
"""

import torch
from torch.optim import AdamW
from transformers import get_linear_schedule_with_warmup


from botiverse.models.TRIPPY.data import prepare_data, Dataset
from botiverse.models.TRIPPY.train import train
from botiverse.models.TRIPPY.evaluate import eval


def run(model, domains, slot_list, label_maps, train_path, dev_path, test_path, device, non_referable_slots, non_referable_pairs, model_path, TRIPPY_config):
    """
    Train and evaluate the TRIPPY model.

    :param model: The TRIPPY model.
    :type model: TRIPPY
    :param domains: The domains to consider in the dataset.
    :type domains: list
    :param slot_list: The list of slots.
    :type slot_list: list
    :param label_maps: The mapping of slot values to their variants.
    :type label_maps: dict
    :param train_path: The path to the training dataset in JSON format.
    :type train_path: str
    :param dev_path: The path to the development dataset in JSON format.
    :type dev_path: str
    :param test_path: The path to the testing dataset in JSON format.
    :type test_path: str
    :param device: The device to train and evaluate the model on.
    :type device: torch.device
    :param non_referable_slots: The slots that are not referable.
    :type non_referable_slots: list
    :param non_referable_pairs: The pairs of slots that are not referable.
    :type non_referable_pairs: list
    :param model_path: The path to save the best model.
    :type model_path: str
    :param TRIPPY_config: The configuration for TRIPPY.
    :type TRIPPY_config: TRIPPYConfig
    """

    n_slots = len(slot_list)

    # train
    print('Preprocessing train set...')
    train_raw_data, train_data = prepare_data(train_path, slot_list, label_maps, TRIPPY_config.tokenizer, TRIPPY_config.max_len, domains, non_referable_slots, non_referable_pairs, TRIPPY_config.multiwoz)
    train_dataset = Dataset(train_data, n_slots, TRIPPY_config.oper2id, slot_list)
    train_sampler = torch.utils.data.RandomSampler(train_dataset)
    train_data_loader = torch.utils.data.DataLoader(train_dataset,
                                                    sampler=train_sampler,
                                                    batch_size=TRIPPY_config.train_batch_size)

    # dev
    if dev_path is not None:
        print('Preprocessing dev set...')
        dev_raw_data, dev_data = prepare_data(dev_path, slot_list, label_maps, TRIPPY_config.tokenizer, TRIPPY_config.max_len, domains, non_referable_slots, non_referable_pairs, TRIPPY_config.multiwoz)
        dev_dataset = Dataset(dev_data, n_slots, TRIPPY_config.oper2id, slot_list)
        dev_data_loader = torch.utils.data.DataLoader(dev_dataset,
                                                    batch_size=TRIPPY_config.dev_batch_size)

    # test
    if test_path is not None:
        print('Preprocessing test set...')
        test_raw_data, test_data = prepare_data(test_path, slot_list, label_maps, TRIPPY_config.tokenizer, TRIPPY_config.max_len, domains, non_referable_slots, non_referable_pairs, TRIPPY_config.multiwoz)
        test_dataset = Dataset(test_data, n_slots, TRIPPY_config.oper2id, slot_list)
        test_data_loader = torch.utils.data.DataLoader(test_dataset,
                                                    batch_size=TRIPPY_config.test_batch_size)

    param_optimizer = list(model.named_parameters())
    no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]
    optimizer_parameters = [
        {
            "params": [
                p for n, p in param_optimizer if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": TRIPPY_config.weight_decay,
        },
        {
            "params": [
                p for n, p in param_optimizer if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]

    # num_train_steps = int(len(train_dataset) / TRAIN_BATCH_SIZE * EPOCHS)
    num_train_steps = len(train_data_loader) * TRIPPY_config.epochs
    num_warmup_steps = int(num_train_steps * TRIPPY_config.warmup_proportion)

    optimizer = AdamW(optimizer_parameters, lr=TRIPPY_config.lr, eps=TRIPPY_config.adam_epsilon)
    scheduler = get_linear_schedule_with_warmup(
        optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=num_train_steps
    )

    best_joint = -1
    for epoch in range(TRIPPY_config.epochs):
        print(f'\nEpoch: {epoch} ---------------------------------------------------------------')
        print('Training the model...')
        train(train_data_loader, model, optimizer, device, scheduler, n_slots, TRIPPY_config.ignore_idx, TRIPPY_config.oper2id)
        if dev_path is not None:
            print('Evaluating the model on dev set...')
            joint_goal_acc, per_slot_acc, macro_f1_score, all_f1_score = eval(dev_raw_data, dev_data, model, device, n_slots, slot_list, label_maps, TRIPPY_config.oper2id, TRIPPY_config.multiwoz)
            print(f'Joint Goal Acc: {joint_goal_acc}')
            print(f'Per Slot Acc: {per_slot_acc}')
            print(f'Macro F1 Score: {macro_f1_score}')
            print(f'All f1 score = {all_f1_score}')
            if joint_goal_acc > best_joint:
                torch.save(model.state_dict(), model_path)
                best_joint = joint_goal_acc
        else:
            torch.save(model.state_dict(), model_path)
    
    if dev_path is not None:
        print('Loading best model on dev set...')

    model.load_state_dict(torch.load(model_path))

    if test_path is not None:
        print('Evaluating the model on test set...')
        joint_goal_acc, per_slot_acc, macro_f1_score, all_f1_score = eval(test_raw_data, test_data, model, device, n_slots, slot_list, label_maps, TRIPPY_config.oper2id, TRIPPY_config.multiwoz)
        print(f'Joint Goal Acc: {joint_goal_acc}')
        print(f'Per Slot Acc: {per_slot_acc}')
        print(f'Macro F1 Score: {macro_f1_score}')
        print(f'All f1 score = {all_f1_score}')


"""
This module contains utility functions for the BERT model.
"""

from collections import OrderedDict
from transformers import BertModel
from botiverse.models.BERT.config import BERTConfig
from botiverse.models.BERT.BERT import Bert


# Load pre-trained weights from trnasformers library
def LoadPretrainedWeights(model):
    """
    Load pre-trained weights from the transformers library.

    This function loads the pre-trained weights from the transformers library
    and updates the model's state_dict accordingly.

    :param model: The BERT model.
    :type model: Bert
    """

    # Get pre-trained weights from transformers library
    pretrained_model = BertModel.from_pretrained('bert-base-uncased')
    state_dict = pretrained_model.state_dict()

    # Delete position_ids from the state_dict if available
    if 'embeddings.position_ids' in state_dict.keys():
        del state_dict['embeddings.position_ids']

    # Get the new weights keys from the model
    new_keys = list(model.state_dict().keys())

    # Get the weights from the state_dict
    old_keys = list(state_dict.keys())
    weights = list(state_dict.values())

    # Create a new state_dict with the new keys
    new_state_dict = OrderedDict()
    for i in range(len(new_keys)):
        new_state_dict[new_keys[i]] = weights[i]
        # print(old_keys[i], '->', new_keys[i])

    model.load_state_dict(new_state_dict)

# Example comparing the outputs of the from scratch model to the pre-trained model from transformers library
import torch
from transformers import BertModel, BertTokenizer

def Example():
    """
    Example comparing the outputs of the from scratch model to the pre-trained model from transformers library.
    """
    
    # Build a BERT model from scratch
    config = BERTConfig()
    model = Bert(config)
    LoadPretrainedWeights(model)

    # Load pre-trained weights from the Transformers library
    pretrained_weights = 'bert-base-uncased'
    pretrained_model = BertModel.from_pretrained(pretrained_weights)

    # Tokenize the input sequence
    tokenizer = BertTokenizer.from_pretrained(pretrained_weights)
    input_text = ["This is a sample input sequence.", "batyousef is awesome"]
    inputs = tokenizer(input_text, padding=True, truncation=True, return_tensors='pt')

    # Set dropout to zero during inference
    model.eval()
    pretrained_model.eval()

    ids = inputs['input_ids']
    token_type_ids = inputs['token_type_ids']
    attention_mask = inputs['attention_mask']

    # Pass the inputs through both models and compare the outputs
    with torch.no_grad():
        model_output1, model_output2 = model(ids, token_type_ids, attention_mask)
        pretrained_output1, pretrained_output2 = pretrained_model(ids,
                                            attention_mask=attention_mask,
                                            token_type_ids=token_type_ids,
                                            return_dict=False
                                            )

    print(model_output1.size(), model_output1)
    print(pretrained_output1.size(), pretrained_output1)
    print()
    print()
    print(model_output2.size(), model_output2)
    print(pretrained_output2.size(), pretrained_output2)

    print(model)
    print(pretrained_model)



from gensim.utils import tokenize
import numpy as np
from nltk.stem.porter import PorterStemmer
stemmer = PorterStemmer()

class TF_IDF():
    '''
    An interface for transforming sentences into tf-idf vectors.
    '''
    def __init__(self):
        self.tf = None
        self.idf = None
        self.all_words = None
    
    def transform_list(self, sentence_list, all_words):
        '''
        Given a list of tokenized sentences, return a table of tf-idf vectors (one for each sentence) in the form of a numpy array.
        :param: sentence_list: A list of tokenized sentences
        :param: all_words: A list of all the words in the corpus
        :return: A numpy array of tf-idf vectors
        '''
        self.all_words = all_words
        # Compute the normalized frequency of each word in the document
        tf_table = np.zeros((len(sentence_list), len(all_words)), dtype=np.float64)
        for i, sentence in enumerate(sentence_list):
            sentence = list(tokenize(sentence, to_lower=True))
            sentence = [stemmer.stem(word.lower()) for word in sentence if word not in ['?', '!', '.', ',']]
            sentence_length = len(sentence)
            for word in sentence:
                word_index = all_words.index(word)
                tf_table[i, word_index] += 1 / sentence_length
        
        # Get the number of documents in which each word appears
        df = np.sum(tf_table > 0, axis=0)
        N = len(sentence_list)
        idf_col = np.log(N/(df+1))
        
        # compute the tf-table by the idf column (gets broadcasted)
        self.tf, self.idf = tf_table, idf_col
        tfidf_table = tf_table * idf_col
        
        return tfidf_table
    

    def transform(self, sentence):
        '''
        Given a sentence, return its tf-idf vector as a numpy array.
        :param: sentence: A string of words
        :return: A numpy array of the tf-idf vector
        '''
        sentence = list(tokenize(sentence, to_lower=True))
        sentence = [stemmer.stem(word.lower()) for word in sentence if word not in ['?', '!', '.', ',']]
        # compute the tf-idf vector for the prompt
        tf_idf = np.zeros((1, len(self.all_words)), dtype=np.float64)
        for word in sentence:
            # get its tf
            if word not in self.all_words:  continue
            word_index = self.all_words.index(word)
            tf_idf[0, word_index] += 1 / len(sentence)
        tf_idf *= self.idf
        return tf_idf
    
    



from transformers import BertTokenizer, BertModel
from botiverse.models.BERT.config import BERTConfig
from botiverse.models.BERT.utils import LoadPretrainedWeights
from botiverse.models import Bert
import torch
import numpy as np


#out = BertEmbedder().embedd(['hello world', 'hello world'])

class BertEmbedder():
    '''An interface for converting given text into BERT embeddings.'''
    def __init__(self):
        '''Load the pre-trained model and tokenizer'''
        self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
        self.bert = Bert(BERTConfig())
        LoadPretrainedWeights(self.bert)
        self.model = BertModel.from_pretrained('bert-base-uncased')
    
    def embed(self, sentences, random_state=42):
        '''
        Convert the given sentences into BERT embeddings.
        :param sentences: A list of sentences to convert into BERT embeddings.
        :return: A list of BERT embeddings for the given sentences.
        '''
        torch.manual_seed(random_state)
        tokss = self.tokenizer(sentences, return_tensors="pt", padding=True, truncation=True)
        emb, clst = self.bert(**tokss)
        attention_mask = tokss['attention_mask']
        # exclude padding tokens
        emb = emb * attention_mask.unsqueeze(-1)
        emb = emb.sum(dim=1) / attention_mask.sum(dim=1, keepdim=True)
        return emb.squeeze()

    
    def closest_sentence(self, new_sentence,  sentence_list, retun_ind=False):
        '''
        Given a list of sentences and a new sentence, return the sentence from the list that is closest to the new sentence.
        :param new_sentence: The new sentence to compare to the list of sentences.
        :param sentence_list: A list of sentences to compare the new sentence to.
        :return: The sentence from the list that is closest to the new sentence and its score.
        '''
        new_sentence_embedding = self.embed(new_sentence)
        sentence_list_embeddings = [self.embed(sentence) for sentence in sentence_list]
        cosine_sim = lambda x, y: torch.dot(x, y) / (torch.norm(x) * torch.norm(y))
        scores = [cosine_sim(new_sentence_embedding, sentence_embedding) for sentence_embedding in sentence_list_embeddings]
        softmax = lambda x: torch.exp(x) / torch.sum(torch.exp(x))
        scores = softmax(torch.tensor(scores)).detach().numpy()
        return sentence_list[np.argmax(scores)] if not retun_ind else np.argmax(scores), np.max(scores)
    
    
try:
    from sentence_transformers import SentenceTransformer, util
except:
    pass

class BertSentenceEmbedder():
    '''
    An interface for converting given text into sentence BERT embeddings.
    '''
    def __init__(self):
        '''
        Load the pre-trained model and tokenizer
        '''
        self.model = SentenceTransformer('all-MiniLM-L6-v2')
        
    def embed(self, sentences):
        '''
        Convert the given sentences into BERT embeddings.
        :param sentences: A list of sentences to convert into BERT embeddings.
        :return: A list of BERT embeddings for the given sentences.
        '''
        return self.model.encode(sentences, convert_to_tensor=True)
    
    def closest_sentence(self, new_sentence,  sentence_list, retun_ind=False):
        '''
        Given a list of sentences and a new sentence, return the sentence from the list that is closest to the new sentence.
        '''
        new_sentence_embedding = self.embed(new_sentence)
        sentence_list_embeddings = self.embed(sentence_list)
        cosine_sim = lambda x, y: torch.dot(x, y) / (torch.norm(x) * torch.norm(y))
        scores = [cosine_sim(new_sentence_embedding, sentence_embedding) for sentence_embedding in sentence_list_embeddings]
        softmax = lambda x: torch.exp(x) / torch.sum(torch.exp(x))
        scores = softmax(torch.tensor(scores)).detach().numpy()        
        return sentence_list[np.argmax(scores)] if not retun_ind else np.argmax(scores), np.max(scores)



try:
    from audiomentations import Compose, AddGaussianNoise, TimeStretch, PitchShift, Shift
    from transformers import Wav2Vec2Model, Wav2Vec2FeatureExtractor, Wav2Vec2Processor, Wav2Vec2ForCTC, Wav2Vec2PhonemeCTCTokenizer
    import os
    import torch
    import torchaudio
    from tqdm import tqdm
    import numpy as np
    # disable warnings from this file
    from transformers import logging
except:
    pass

class Wav2Vec():
    '''
    An interface for transforming audio files into wav2vec vectors.
    '''
    def __init__(self, sample_rate=16000, duration=1, augment= None):
        '''
        Initialize the Wav2Vec transformer by loading the wav2vec model and setting the sample rate and duration of the audio files.
        :param: sample_rate: The sample rate of the audio files
        :param: duration: The duration of the audio files in milliseconds        
        '''
        logging.set_verbosity_error()
        self.model = Wav2Vec2Model.from_pretrained("facebook/wav2vec2-base-960h")
        self.extractor = Wav2Vec2FeatureExtractor.from_pretrained("facebook/wav2vec2-base-960h")

        self.sample_rate = sample_rate
        self.duration = duration
        self.emb_dim = 768 * int(49 * self.duration)         # fact regarding wav2vec2-base-960h
        if augment is None:
            self.augment = Compose([
            TimeStretch(min_rate=0.8, max_rate=1.4, p=0.7),
            PitchShift(min_semitones=-4, max_semitones=1, p=0.8),
            Shift(min_fraction=-0.5, max_fraction=0.5, p=0.5),
            AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.015, p=0.5),
        ])
        else:
            self.augment = augment
            
    def transform_list(self, words, n=4):
        '''
        Given a folder dataset with folders each containing audio files, this returns a table of wav2vec vectors (one for each audio file) in the form of a numpy array X and a table of classes in the form of a numpy array y.
        Note that in the process, each audio file is augmented n times and each corresponds to another wav2vec vector.
        '''
        sounds_per_word = len(os.listdir(f"dataset/{words[0]}"))
        self.N = len(words) * sounds_per_word
        X, y = [], []
        print("Transforming audio files into embeddings...")
        for word in tqdm(words):
            for i, file in enumerate(os.listdir(f"dataset/{word}")):
                if not file.endswith(".wav"): continue
                waveform, sr = torchaudio.load(f"dataset/{word}/{file}")
                # resample if sr != 24K
                waveform = torchaudio.transforms.Resample(sr, self.sample_rate)(waveform)
                
                if waveform.shape[0] == 2:
                    waveform = torch.mean(waveform, dim=0, keepdim=True)
                
                length = waveform.shape[1]
                
                sample_dur = int(self.duration * self.sample_rate)
                if length < sample_dur:
                    waveform = torch.cat((waveform, torch.zeros(1, sample_dur - length)), dim=1)
                elif length > sample_dur:
                    waveform = waveform[:, :sample_dur]         

                waveform = waveform.squeeze()
                waveform = waveform.detach().numpy()

                if n >0:
                    for _ in range(n):
                        waveform = self.augment(samples=waveform, sample_rate=self.sample_rate)
                        inputs = self.extractor(waveform, return_tensors="pt", padding=True, sampling_rate=self.sample_rate)
                        features = self.model(inputs.input_values).last_hidden_state
                        features = features.squeeze().detach().numpy()
                        X.append(features)
                        y.append(words.index(word))
                else:
                    inputs = self.extractor(waveform, return_tensors="pt", padding=True, sampling_rate=self.sample_rate)
                    features = self.model(inputs.input_values).last_hidden_state
                    features = features.squeeze().detach().numpy()
                    X.append(features)
                    y.append(words.index(word))
        X = np.array(X)
        y = np.array(y)
        return X, y
    
    def transform(self, path, strict_duration=False):
        '''
        Convert the audio file as in the path into a wav2vec vector.
        :param: path: The path to the audio file
        :param: strict_duration: If True, the audio file is padded or truncated to the duration specified during init.
        '''
        waveform, sr = torchaudio.load(path)
        waveform = torchaudio.transforms.Resample(sr, self.sample_rate)(waveform)
        if waveform.shape[0] == 2:  waveform = torch.mean(waveform, dim=0, keepdim=True)
        length = waveform.shape[1]
        if strict_duration:
            sample_dur = int(self.duration * self.sample_rate)
            if length < sample_dur:
                waveform = torch.cat((waveform, torch.zeros(1, sample_dur - length)), dim=1)
            elif length > sample_dur:
                waveform = waveform[:, :sample_dur]    
        waveform = waveform.squeeze()
        inputs = self.extractor(waveform, return_tensors="pt", padding=True, sampling_rate=self.sample_rate)
        features = self.model(inputs.input_values).last_hidden_state.detach().numpy()
        return features
    
    

from transformers import Wav2Vec2Tokenizer, Wav2Vec2ForCTC
from transformers import logging

# Load the pre-trained model and tokenizer

class Wav2Text():
    ''' An interface for converting speech files into text using wav2vec2.'''
    def __init__(self):
        '''Load the pre-trained model and tokenizer'''
        logging.set_verbosity_error()
        self.model = Wav2Vec2ForCTC.from_pretrained("facebook/wav2vec2-base-960h")
        self.tokenizer = Wav2Vec2Tokenizer.from_pretrained("facebook/wav2vec2-base-960h")
        
    def transcribe(self, path):
        '''
        Given a path to a speech file, return the transcription of the speech file.
        :param: path: The path to the speech wav file
        :return: The transcription of the speech file
        '''
        # load audio
        waveform, sample_rate = torchaudio.load(path)
        
        # resample if sr not 16K
        resampler = torchaudio.transforms.Resample(sample_rate, 16000)
        waveform = resampler(waveform)
        waveform = waveform.squeeze()

        # preprocess the audio
        input_values = self.tokenizer(waveform, return_tensors="pt").input_values

        # Perform speech-to-text conversion
        with torch.no_grad():
            logits = self.model(input_values).logits

        # Decode the predicted transcription
        predicted_ids = torch.argmax(logits, dim=2)
        transcription = self.tokenizer.batch_decode(predicted_ids)[0]
        return transcription.lower()




"""
This module ontains utility code used by the basic Taskbot module
such as Natural Language Understanding (NLU), Dialogue State Tracker (DST) ... etc.
"""
import re
import random

class RuleBasedNLU:
    """
    Instantiate a Rule-Based Natural Language Understanding module.
    It uses Regex to extract domain and slot-fillers from the user's utterance.

    :param domains_pattern: Dictionary where the key is a name of a domain
        and the value is a Regex that is used to capture that domain.
    :type domains_pattern: dict[str, str]

    :param slots_pattern: Dictionary where the key is a name of a domain
        and the value is another dictionary where the key is a name of a slot
        inside the domain and the value is a Regex that is used to capture that slot.
    :type slots_pattern: dict[str, dict[str, str]]
    """

    def __init__(self, domains_pattern, slots_pattern):
        self.domains_pattern = domains_pattern
        self.slots_pattern = slots_pattern

    def get_domain(self, prev_domain, utterance):
        """
        Takes previously extracted domain and the user's utterance and extract the domain from it,
        if no domain can be extracted it returns prev_domain.

        :param prev_domain: The domain from previous user's utterance.
        :type prev_domain: str

        :param utterance: The user's utterance.
        :type utterance: str

        :return: The success of the domain extraction & The extracted domain.
        :rtype: tuple[bool, str]
        """
        detected = []
        for domain, pattern in self.domains_pattern.items():
            if re.search(pattern, utterance):
                detected.append(domain)

        if len(detected) > 1:
            return False, ""

        if len(detected) == 0:
            return True, prev_domain

        return True, detected[0]

    def get_slot_fillers(self, current_domain, utterance):
        """
        Takes the current extracted domain and the user's utterance
        and extract the slot-fillers.

        :param current_domain: The domain extracted from the current
            user's utterance.
        :type utterance: str

        :param utterance: The user's utterance.
        :type utterance: str

        :return: The success of the slot-fillers extraction, the extracted slots
            & the extracted slot-values.
        :rtype: tuple[bool, list[str], list[str]]
        """
        slots, values = [], []
        for slot, pattern in self.slots_pattern[current_domain].items():
            matches = set()
            for match in re.finditer(pattern, utterance):
                matches.add(" ".join(match.groups()))

            if len(matches) > 1:
                return False, [], []

            if len(matches) == 1:
                slots.append(slot)
                values.append(list(matches)[0])

        return True, slots, values

class MostRecentDST:
    """
    Instantiate a Dialogue State Tracker module.
    It maintains the state of the dialogue which is mainly represented in
    the slots and the slots-values that are extracted from the user's utterances.
    It is a most recent tracker which means it only keeps track of the recent value in case
    if multiple values appear for the same slot.

    :param domains_slots: The slots of each domain in the system.
    :type domains_slots: dict[str, list[str]]
    """
    def __init__(self, domains_slots):
        self.state = {}
        for domain, slots in domains_slots.items():
            self.state[domain] = {}
            for slot in slots:
                self.state[domain][slot] = None

    def update_state(self, domain, slots, values):
        """
        Update the slots values of a certain domain.

        :param domain: The domain of the slots to be updated.
        :type domain: str

        :param slots: List the of the slots to be updated.
        :type slots: list[str]

        :param values: List of the values of the slots.
        :type values: list[str]
        """
        for i, slot in enumerate(slots):
            self.state[domain][slot] = values[i]

    def get_dialogue_state(self):
        """
        Gets the dialogue state.

        :return: The dialogue state, where state["domain"]["slot"] indicates
            value of slot "slot" in domain "domain".
        :rtype: dict[str, dict[str, str]]
        """
        return self.state

    def reset(self, domain=None, slot=None):
        """
        Reset the dialogue state.
        Note: if the domain or the slot is equal to None all domains
        or slots will be reset respectively.

        :param domain: The domain to be reset, defaults to None.
        :type domain: str

        :param slot: The slot to be reset, defaults to None.
        :type slot: str

        """
        if domain is None and slot is None:
            for cur_domain, slot_filler in self.state.items():
                for cur_slot in slot_filler:
                    self.state[cur_domain][cur_slot] = None
        elif domain is not None and slot is None:
            for cur_slot in self.state[domain]:
                self.state[domain][cur_slot] = None
        elif domain is None and slot is not None:
            for cur_domain, slot_filler in self.state.items():
                if slot in slot_filler:
                    self.state[cur_domain][slot] = None
        else:
            self.state[domain][slot] = None

class RandomDP:
    """
    Instantiate a Random Dialogue Policy module.
    Randomly determines the next action which is the
    next empty slot to ask user about.
    """
    def get_action(self, current_domain, state):
        """
        Takes the current domain and dialogue state and return
        the next action which is the next empty slot to ask user about.

        :param current_domain: The current extracted domain.
        :type current_domain: str

        :param state: The current state of the dialogue.
        :type state: dict[str, dict[str, str]]

        :return: The next empty slot to ask the user about.
        :rtype: str
        """

        unfilled = []
        for slot, value in state[current_domain].items():
            if value is None:
                unfilled.append(slot)

        max_len = len(unfilled)

        if max_len == 0:
            return ""

        return unfilled[random.randint(0, max_len-1)]

class TemplateBasedNLG:
    """
    Instantiate a Template Based Natural Language Generation module.
    It uses a predefined templates to generate a response to the user
    and ask about empty slots.

    :param templates: The predefined templates, where templates["domain"]["slot"] is
        a list of questions asking about slot "slot" in domain "domain".
    :type templates: dict[str, dict[str, list[str]]]
    """
    def __init__(self, templates):
        self.templates = templates

    def generate(self, domain, slot):
        """
        Takes a slot and the domain of the slot and generate a question randomly
        from the templates about the slot.

        :param domain: The domain of the slost.
        :type domain: str

        :param slot: The slot.
        :type slot: str

        :return: The generated question.
        :rtype: str
        """
        max_len = len(self.templates[domain][slot])
        return self.templates[domain][slot][random.randint(0, max_len-1)]



import torch
from transformers import AutoModelForSeq2SeqLM
from tqdm.auto import tqdm
import torch.optim as optim
from botiverse.preprocessors.Special.ConverseBot_Preprocessor.ConverseBot_Preprocessor import ConverseBot_Preprocessor
from botiverse.models.T5Model.T5Model import T5Model

class ConverseBot:
    '''An interface for the ConverseBot model which is a conversational model based on the Flan-T5 model'''
    def __init__(self, from_scratch=False):
        """
        Initializes a ConverseBot instance and loads the Backend finetuning parameters, and optionally gets the training dataset if a frontend finetuning is desired.

        :param from_scratch: Boolean flag to indicate whether to load a model verstion that is made from scratch (recommended to be False)
        :type from_scratch: Boolean, optional

        :returns: None
        """
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.from_scratch = from_scratch
        if self.from_scratch:
            # create a model instance
            self.model = T5Model()
            # load the Backend finetuning parameters
            self.model.load_state_dict(AutoModelForSeq2SeqLM.from_pretrained("MohamedSaad/T5_ConverseBot").state_dict())
        else:
            self.model = AutoModelForSeq2SeqLM.from_pretrained("MohamedSaad/T5_ConverseBot")
        # move the model to the GPU if available
        self.model.to(self.device)
        # load the preprocessor
        self.preprocessor = ConverseBot_Preprocessor()

    def read_data(self, file_path):
        """
        Reads and pre-processes the data, sets up the model based on the data and prepares the train-validation split.

        :param file_path: The path to the file that contains the dataset.
        :type file_path: str

        :returns: None
        """
        # read data
        self.preprocessor = ConverseBot_Preprocessor(file_path=file_path)
        # process data
        self.data = self.preprocessor.process()
        # train validation split
        self.train_data = self.data.sample(frac=0.99, random_state=0)
        self.validation_data = self.data.drop(self.train_data.index)
        self.train_data = self.train_data.reset_index(drop=True)
        self.validation_data = self.validation_data.reset_index(drop=True)

    def train(self, epochs=1, batch_size=1):
        """
        Trains the model on the input dataset.

        :param epochs: Number of epochs to train for.
        :type epochs: int, optional

        :param batch_size: The size of the training batches.
        :type batch_size: int, optional

        :returns: None
        """
        self.model.train()
        self.optimizer = optim.Adam(self.model.parameters(), lr=0.00005)
        for epoch in range(epochs):
            for i in tqdm(range(0, len(self.train_data), batch_size)):
                self.model.zero_grad()
                # prepare the training batches
                batch_text_input_ids = torch.concat(self.train_data['text_input_ids'][i:i+batch_size].tolist()).to(self.device)
                batch_text_attention_mask = torch.concat(self.train_data['text_attention_mask'][i:i+batch_size].tolist()).to(self.device)
                batch_labels = torch.concat(self.train_data['target'][i:i+batch_size].tolist()).to(self.device).to(self.device)
                if self.from_scratch:
                    loss = self.model(input_ids=batch_text_input_ids, attention_mask=batch_text_attention_mask[0], decoder_input_ids=batch_labels)[1]
                else:
                    loss = self.model(input_ids=batch_text_input_ids, attention_mask=batch_text_attention_mask, labels=batch_labels).loss
                loss.backward()
                self.optimizer.step()
            print("Epoch: " + str(epoch) + " Loss: " + str(loss.item()))

    def validation(self, batch_size=1):
        """
        Validates the model on the validation dataset.

        :param batch_size: The size of the validation batches.
        :type batch_size: int, optional

        :returns: None
        """
        total = 0
        loss = 0
        self.model.eval()
        with torch.no_grad():
            for i in tqdm(range(0, len(self.validation_data), batch_size)):
                # prepare the validation batches
                batch_text_input_ids = torch.concat(self.validation_data['text_input_ids'][i:i+batch_size].tolist()).to(self.device)
                batch_text_attention_mask = torch.concat(self.validation_data['text_attention_mask'][i:i+batch_size].tolist()).to(self.device)
                batch_labels = torch.concat(self.validation_data['target'][i:i+batch_size].tolist()).to(self.device)
                outputs = self.model(input_ids=batch_text_input_ids, attention_mask=batch_text_attention_mask, labels=batch_labels)
                loss += outputs.loss.item()
                total += batch_labels.size(0)
        print('Validation Loss: ', loss/total)

    def infer(self, string, temperature=1):
        """
        Inference on the model using the input string.

        :param string: The string to provide for inference.
        :type string: str

        :param temperature: The temperature of the softmax function, the higher its value the flatter the probability distribution of the next token will be.
        :type temperature: float, optional

        :returns: Inference result from the model.
        :rtype: str
        """
        self.model.eval()
        token_obj = self.preprocessor.process_string(string)
        input_ids= token_obj['input_ids'].to(self.device)
        attention_mask = token_obj['attention_mask'].to(self.device)
        if self.from_scratch:
            output_tokens = self.model.generate(input_ids=input_ids, attention_mask=attention_mask, max_length=250, temperature=temperature)
            return self.preprocessor.decode_tokens(output_tokens)
        else:
            output_tokens = self.model.generate(input_ids=input_ids, attention_mask=attention_mask, max_length=250, num_beams=5, early_stopping=True, temperature=temperature, repetition_penalty=1.5)
            return self.preprocessor.decode_tokens(output_tokens[0])

    # save a model locally
    def save(self, path):
        """
        Save the model locally to the provided path.

        :param path: The path where to save the model.
        :type path: str

        :returns: None
        """
        torch.save(self.model.state_dict(), path)

    # load a model from a local path
    def load(self, path):
        """
        Load the model from the provided path.

        :param path: The path where to load the model from.
        :type path: str

        :returns: None
        """
        self.model.load_state_dict(torch.load(path))


try:
    import pyaudio
except:
    pass
import wave
from array import array
from tqdm import tqdm

def voice_input(record_time=3, voice_threshold=900, save_path='sample.wav'):
    '''
    Upon call, record audio for record_time seconds and save it to save_path while only inputting audio that is above the voice_threshold.
    :param record_time: The number of seconds to record for.
    :param voice_threshold: The minimum volume of audio to record.
    :param save_path: The path to save the audio file to.
    :return: The path to the audio file.
    '''
    # """"
    #instantiate the pyaudio
    audio = pyaudio.PyAudio()

    #recording prerequisites
    chunk, channels, sample_rate = 1024, 1, 16000
    stream = audio.open(format=pyaudio.paInt16, channels=channels, rate=sample_rate, input=True, frames_per_buffer=1024)

    #starting recording
    frames=[]
    pbar = tqdm(range(0,int(sample_rate/chunk*record_time)), desc=f"Recording")
    for i in pbar:
        data = stream.read(chunk)
        data_chunk = array('h', data)
        chunk_volume = max(data_chunk)
        if voice_threshold is not None:
            state = 'Talking' if chunk_volume > voice_threshold else 'Silence'
            if state == 'Talking': frames.append(data)
        else:
            frames.append(data)
            state = 'Recording'
        pbar.set_postfix({"Volume": chunk_volume, "State": state})

    stream.stop_stream(); stream.close(); audio.terminate()
    
    # save audio recorded from stream
    wavfile = wave.open(save_path,'wb')
    wavfile.setnchannels(channels)
    wavfile.setsampwidth(audio.get_sample_size(pyaudio.paInt16))
    wavfile.setframerate(sample_rate)
    wavfile.writeframes(b''.join(frames))
    wavfile.close()

    return save_path
    # """"
    #pass


import os
import gdown
from gensim.utils import tokenize
import numpy as np


class GloVe():
    '''
    An interface for transforming words into GloVe vectors.
    '''
    def __init__(self, force_download=False):
        '''
        Initialize the GloVe transformer and download the embeddings if needed.
        :param: force_download: If True, download the embeddings even if they already exist.
        '''
        self.glove_dict = None
        self.load_glove_vectors(force_download=force_download)

    # Load GLOVE word vectors
    def load_glove_vectors(self, force_download):
        '''
        Load GloVe vectors from gensim into the class. By default uses 50D vectors.
        :param: force_download: If True, download the embeddings even if they already exist.
        '''
        curr_dir = os.path.dirname(os.path.abspath(__file__))
        path = curr_dir + '/glove.6B.50d.txt'
        if not os.path.exists(path) or force_download:
            print("GLoVE embeddings not found. Downloading now...")
            f_id = '1BOSO0rR3ZzjWlv5WYzCux6ZluBP_vNDv' 
            gdown.download(f'https://drive.google.com/uc?export=download&confirm=pbef&id={f_id}', curr_dir + '/glove.6B.50d.txt', quiet=False)
            print("Done.")   
        
        glove_dict = {}         # dictionary mapping words to their GloVe vector representation
        with open(path, 'r', encoding='utf-8') as f:
            for line in f:
                values = line.split()
                # In each line, the first value is the word, the rest are the values of the vector
                word = values[0]            
                vector = np.asarray(values[1:], dtype='float32')
                glove_dict[word] = vector
        self.glove_dict = glove_dict


    def transform_list(self, sentence_list, **kwargs):
        '''
        Transform a sentence list into a numpy array of GloVe vectors
        :param: sentence_list: A list of sentences
        :return: A numpy array of GloVe vectors
        '''
        # make a numpy array of the sentence vectors
        X = np.zeros((len(sentence_list), 50))
        for i, sentence in enumerate(sentence_list):
                X[i] = self.transform(sentence)
        return X
    
    def transform(self, sentence):
        '''
        Transform a sentence into a GloVe vector by averaging the word vectors in it.
        :param: sentence: a string of words
        :return: the corresponding GloVe vector
        '''
        tokens = list(tokenize(sentence, to_lower=True))
        vector_sum = np.zeros(50)  
        num_glove_tokens = 0            # num of words that occur in glove vocab
        for token in tokens:
            if token in self.glove_dict:
                vector_sum += self.glove_dict[token]
                num_glove_tokens += 1

        avg_vector = np.zeros_like(self.glove_dict['a'])  if num_glove_tokens == 0 else vector_sum / num_glove_tokens
        avg_vector = avg_vector[np.newaxis, :]
        return avg_vector


#!/usr/bin/env python
# coding: utf-8

# # <font color="cyan">Feedforward Neural Network </font> From Scratch Implementation
# In this notebook, we will implement a feedforward neural network with arbitrary structure, loss and activations. You can find a full tutorial on this notebook [here](hhttps://medium.com/towards-data-science/backpropagation-the-natural-proof-946c5abf63b1).

# In[1]:


'''
This module implements and provides an interface for an arbitrary neural network architecture.
'''
try:
    from botiverse.models.NN.utils import split_data, batchify
except:
    pass
# check if running on notebook
from tqdm import tqdm
import numpy as np
import copy 
import pickle


# ### Define the Architecture

# In[2]:


class NeuralNet():
    '''
    Defines the hypothesis set and learning algorithm for a neural network.
    '''
    def __init__(self, structure, activation='sigmoid', optimizer='ADAM'):
        '''
        Initialize the hyperparameters, weights and biases for the network. 
        :param structure: A list of integers representing the number of neurons in each layer. For example, [2, 3, 1] is a network with 2 neurons in the input layer, 3 in the hidden layer, and 1 in the output layer.
        :param activation: The activation function to use. Can be 'sigmoid' or 'relu'.
        :param optimizer: The optimizer to use. Can be 'ADAM' or 'SGD'.
        '''
        ### Hyperparameters
        self.structure = structure
        self.num_layers = len(structure) 
        self.activation = activation
        self.optimizer = optimizer
        if activation == 'sigmoid':
            def σ(z): return 1.0/(1.0+np.exp(-z))                   # activation function
            def σࠤ(z): return σ(z)*(1-σ(z))                         # derivative of the activation function
     
        elif activation == 'relu':
            def σ(z): return np.maximum(0, z)
            def σࠤ(z): return np.greater(z, 0).astype(int)
        self.h = σ
        self.dh = σࠤ   
        
        def softmax(z): return np.exp(z-np.max(z))/np.sum(np.exp(z-np.max(z)), axis=0)
        self.g = softmax
        
        ### Parameters
        # Xaiver initialization
        #for each layer except the first y neurons randomize a (y, 1) vector for the bias vector:
        self.Bₙ = [np.random.randn(l, 1) * np.sqrt(2/l) for l in structure[1:]]
        #for each two consecutive layers with x and y neurons respectively randomize a (y,x) matrix for the weight matrix:
        self.Wₙ = [np.random.randn(l, next_l) * np.sqrt(2/l) for l, next_l in zip(structure[:-1], structure[1:])]
        
        ### Loss
        def cross_entropy(aᴺ, y): return -np.sum(y * np.log(aᴺ)) / aᴺ.shape[1]
        def cross_entropyࠤ(aᴺ, y): return (aᴺ-y)
        self.J = cross_entropy
        self.ᐁJ = cross_entropyࠤ

        ### Misc
        self.optimal_epochs = None
NNClass = lambda func: setattr(NeuralNet, func.__name__, func) or func


# ### Backpropagation

# In[3]:


@NNClass
def backprop(self, xₛ , yₛ ):
    '''
    Compute the loss gradients მJⳆმBₙₛ, მJⳆმWₙₛ given an observation (xₛ , yₛ ) where xₛ and yₛ are column vectors. 
    :param xₛ: The input vector of shape (d, 1).
    :param yₛ: The output vector of shape (K, 1).
    :return: A tuple of lists (მJⳆმBₙₛ, მJⳆმWₙₛ) where მJⳆმBₙₛ is a list of the gradients of the loss function with respect to the biases and მJⳆმWₙₛ is a list of the gradients of the loss function with respect to the weights.
    '''
    h, dh, ᐁJ, g = self.h, self.dh, self.ᐁJ, self.g

    მJⳆმBₙₛ = [np.zeros(b.shape) for b in self.Bₙ]
    მJⳆმWₙₛ = [np.zeros(W.shape) for W in self.Wₙ]

    # forward pass (computing z for all layers)
    Zₙ = []                     # list to store all the z vectors, layer by layer
    Aₙ = []                     # list to store all the a vectors layer by layer

    for i, (b, W) in enumerate(zip(self.Bₙ, self.Wₙ)):
        z = W.T @ a + b if Zₙ else W.T @ xₛ  + b
        a = h(z) if i != self.num_layers-2 else g(z)
        Zₙ.append(z)
        Aₙ.append(a)

    #Zₙ and Aₙ are now ready.

    # backward pass (computing δ and consequently მJⳆმBₙₛ and მJⳆმWₙₛ layer by layer )
    H = self.num_layers-2
    for L in range(H, -1, -1):
        δ =  dh(Zₙ[L]) * (self.Wₙ[L+1] @ δ) if L != H else ᐁJ(Aₙ[L], yₛ ) 
        მJⳆმBₙₛ[L] = δ
        მJⳆმWₙₛ[L] = Aₙ[L-1] @ δ.T  if L != 0 else xₛ  @ δ.T
    
    return (მJⳆმBₙₛ, მJⳆმWₙₛ)


# ### Gradient Descent

# In[4]:


@NNClass
def SGD(self, x_batch, y_batch, λ, α=0.01):
    '''
    Given a minibatch (a list/numpy array of tuples (xₛ, yₛ )) this will update Bₙ and Wₙ by applying SGD with L2 regularization.
    :param x_batch: A list/numpy array of input vectors of shape (d, 1).
    :param y_batch: A list/numpy array of output vectors of shape (K, 1).
    :param λ: The learning rate.
    :param α: The regularization parameter.
    :return: None
    '''
    მJⳆმBₙ = [np.zeros(b.shape) for b in self.Bₙ]
    მJⳆმWₙ = [np.zeros(W.shape) for W in self.Wₙ]

    for x, y in zip(x_batch, y_batch):
        მJⳆმBₙₛ, მJⳆმWₙₛ = self.backprop(x, y)
        მJⳆმBₙ = [მJⳆმb + მJⳆმbₛ for მJⳆმb, მJⳆმbₛ in zip(მJⳆმBₙ, მJⳆმBₙₛ)]  
        მJⳆმWₙ = [მJⳆმW + მJⳆმWₛ for მJⳆმW, მJⳆმWₛ in zip(მJⳆმWₙ, მJⳆმWₙₛ)]

    d = len(x_batch)
    self.Wₙ = [(1 - λ * α / d) * W - λ / d * მJⳆმW / np.linalg.norm(მJⳆმW) for W, მJⳆმW in zip(self.Wₙ, მJⳆმWₙ)]
    self.Bₙ = [(1 - λ * α / d) * b - λ / d * მJⳆმb / np.linalg.norm(მJⳆმb) for b, მJⳆმb in zip(self.Bₙ, მJⳆმBₙ)]


# In[5]:


@NNClass
def ADAM(self, x_batch, y_batch, λ, α=0.01, beta1=0.9, beta2=0.999, epsilon=1e-8):
    '''
    Given a minibatch (a list of tuples (xₛ, yₛ )) this will update Bₙ and Wₙ by applying Adam optimizer with L2 regularization.
    :param x_batch: A list/numpy array of input vectors of shape (d, 1).
    :param y_batch: A list/numpy array of output vectors of shape (K, 1).
    :param λ: The learning rate.
    :param α: The regularization parameter.
    :param beta1: The exponential decay rate for the first moment estimates.
    :param beta2: The exponential decay rate for the second-moment estimates.
    :param epsilon: A small constant for numerical stability.
    '''
    mB = [np.zeros(b.shape) for b in self.Bₙ]           # momentum for Bₙ
    vB = [np.zeros(b.shape) for b in self.Bₙ]           # RMSprop for Bₙ
    mW = [np.zeros(W.shape) for W in self.Wₙ]           # momentum for Wₙ
    vW = [np.zeros(W.shape) for W in self.Wₙ]           # RMSprop for Wₙ

    for x, y in zip(x_batch, y_batch):
        mJdB, mJdW = self.backprop(x, y)
        # update the momentum and RMSprop for Wₙ and Bₙ
        mB = [beta1 * mb + (1 - beta1) * mJdb for mb, mJdb in zip(mB, mJdB)]
        vB = [beta2 * vb + (1 - beta2) * (mJdb ** 2) for vb, mJdb in zip(vB, mJdB)]
        mW = [beta1 * mw + (1 - beta1) * mJdw for mw, mJdw in zip(mW, mJdW)]
        vW = [beta2 * vw + (1 - beta2) * (mJdw ** 2) for vw, mJdw in zip(vW, mJdW)]

    # update the parameters Wₙ and Bₙ
    d = len(x_batch)
    self.Wₙ = [(1 - α * λ / d) * W - λ * mb / (np.sqrt(vb) + epsilon) for W, mb, vb in zip(self.Wₙ, mW, vW)]
    self.Bₙ = [(1 - α * λ / d) * b - λ * mb / (np.sqrt(vb) + epsilon) for b, mb, vb in zip(self.Bₙ, mB, vB)]


# ### Feedforward

# In[6]:


@NNClass
def feedforward(self, x ):
    h, g = self.h, self.g
    '''
    The forward pass of the network. Given an input x this will return the output of the network.
    :param x: The input vector of shape (d, 1) where d is the number of input features.
    :return: The output vector of shape (K, 1) where K is the number of classes.
    '''
    a = x
    for i, (b, W) in enumerate(zip(self.Bₙ, self.Wₙ)):
        z = W.T @ a + b
        a = h(z) if i != self.num_layers-2 else g(z)
    ŷ = a
    return ŷ 


# ### Training Method

# In[7]:


@NNClass
def fit(self, X, y, batch_size=32, epochs=200, λ=30, α=0.01, optimizer='ADAM', val_split=0.0, patience=50, eval_train=False):
    '''
    For each epoch, go over each minibatch and perform a gradient descent update accordingly and evaluate the model on the training and validation sets if needed.
    :param X: The training data arranged as a float numpy array of shape (N, d)
    :param y: The training labels arranged as a numpy array of shape (N,) where each element is an integer between 0 and k-1
    :param batch_size: The size of the minibatches to use.
    :param epochs: The number of epochs to run.
    :param λ: The learning rate.
    :param α: The regularization parameter.
    :param optimizer: The optimizer to use. Can be 'ADAM' or 'SGD'.
    :param val_split: The ratio of the validation set to the training set. If given, per-epoch validation will be performed.
    :param eval_train: Whether to evaluate the model on the training set.
    :return: None
    '''
    if val_split:
        X, y, X_v, y_v = split_data(X, y, val_split)
    
    Xc, yc = copy.deepcopy(X), copy.deepcopy(y)
    X, y = batchify(X, y, batch_size)

    if self.optimal_epochs is not None:
        print(f"Using optimal epochs: {self.optimal_epochs} found from an earlier run.")
        epochs = self.optimal_epochs
        
    # training loop
    self.gradient_descent = self.SGD if optimizer == 'SGD' else self.ADAM
    train_acc, val_acc = '', ''
    pbar = tqdm(range(epochs))
    best_loss = np.inf
    val_loss = 0
    bad_epochs = 0
    for epoch in pbar:
        # randomly shuffle X, y
        p = np.random.permutation(len(X))
        X, y = X[p], y[p]
        for x_batch, y_batch in zip(X, y):
            self.gradient_descent(x_batch, y_batch, λ)          #update the parameters after learning from the mini_batch.
        if eval_train:    
            train_acc = self.evaluate(Xc, yc)
    
        # update bar
        desc1, desc2 = f"Train Acc: {train_acc}" if eval_train else '', f" | Val Acc: {val_acc}" if val_split else ''
        desc3 = f" | Val Loss: {val_loss}" if val_split else ''
        desc =  desc1 + desc2 + desc3
        pbar.set_description(desc)
        
        # early stopping
        if val_split:     
            val_loss = self.evaluate(X_v, y_v, loss=True)
            if val_loss <= best_loss:
                best_loss = val_loss
                bad_epochs = 0
            else:
                bad_epochs += 1
                if bad_epochs == patience:
                    print(f"{patience} epochs have passed without improvement. Early stopping... \n")
                    self.optimal_epochs = epoch - patience + 2
                    self.optimal_epochs = int(self.optimal_epochs * (1 + val_split))
                    break
            val_acc = self.evaluate(X_v, y_v)        


# In[8]:


@NNClass
def predict(self, X):
    '''
    Predict the class of each example in X.
    :param X: The test data arranged as a float numpy array of shape (N, d)
    '''
    if len(X.shape) == 2:   X = X[..., np.newaxis]
    # return the predicted class and the probability of the predicted class
    return np.array([np.argmax(self.feedforward(x)) for x in X]), np.array([np.max(self.feedforward(x)) for x in X])
        
@NNClass
def evaluate(self, X,y, loss=False):
    '''
    Compare the one-hot vector y with the networks output yᴺ and calculate the accuracy.
    :param X: The test data arranged as a float numpy array of shape (N, d)
    :param y: The test labels arranged as a numpy array of shape (N,) where each element is an integer between 0 and k-1
    '''
    if len(X.shape) == 2:   X = X[..., np.newaxis]
    if loss:
        return np.sum([self.J(self.feedforward(x), y) for x, y in zip(X, y)]) / len(X)
    else:
        validation_results = [(np.argmax(self.feedforward(x)), y) for (x, y) in zip(X, y)]   #the index is the number itself
        accuracy = sum(int(ŷ == y) for (ŷ, y) in validation_results) / len(X)
    
    return round(accuracy, 2)


# #### Saving and Loading Model

# In[9]:


@NNClass
def save(self, path):
    '''
    Save the model to the given path.
    :param path: The path to save the model to.
    '''
    # store self.Bₙ, self.Wₙ, self.structure, self.activation, self.optimizer
    with open(path, 'wb') as f:
        # save self.Bₙ, self.Wₙ, self.structure, self.activation, self.optimizer
        class_dict = {key: value for key, value in self.__dict__.items() if key in ['Bn', 'Wn', 'structure', 'activation', 'optimizer']}
        f.write(pickle.dumps(class_dict))

@staticmethod        
@NNClass
def load(path):
    '''
    Load the model from the given path.
    :param path: The path to load the model from.
    '''
    with open(path, 'rb') as f:
        class_dict = pickle.loads(f.read())
        model = NeuralNet(class_dict['structure'], class_dict['activation'], class_dict['optimizer'])
        # now use setattr to set Bₙ and Wₙ
        for key, value in class_dict.items():
            if key in ['Bn', 'Wn']:
                setattr(model, key, value)
        return model


# In[10]:


# if running from notebook
if __name__ == '__main__':
    get_ipython().system('jupyter nbconvert --to script NN.ipynb')




import pandas as pd
from transformers import AutoTokenizer
import numpy as np
import json

class ConverseBot_Preprocessor:
    ''''An interface that provides the required preprocessing for the ConverseBot bot'''
    def __init__(self, file_path=None, dataset=None):
        """
        Initializes a ConverseBot_Preprocessor instance with an optional training dataset, note that the dataset structure is an array of multiturn conversations and each multiturn conversation is an array of strings, e.g., [["hi","hello","how are you?"], ["good","how about you?","i am fine"]]

        :param dataset: Dataset to be processed (use it or file_path).
        :type dataset: list of list of str, optional

        :param file_path: Path to the .json file that contains the conversation array (use it or dataset).
        :type file_path: str, optional

        :returns: None
        """
        if file_path:
            with open(file_path, 'r') as f:
                self.data = json.load(f)
        else:
            self.data = dataset
        # create the t5 tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained("google/flan-t5-base")

    def process(self):
        """
        Processes the conversations dataset by cleaning it then combining each conversation into a single string (with [C] between each turn) and then tokenizing it.

        :returns: DataFrame containing the processed conversations.
        :rtype: DataFrame
        """
        # separate conversations into text and target
        target_array = []
        text_array = []
        for conversation in self.data:
            # get length of conversation
            conversation_length = len(conversation)
            # get a random index to split the conversation
            split_index = np.random.randint(0, conversation_length//2)*2+1
            # split the conversation into input and target
            conternt = conversation[:split_index]
            target = conversation[split_index]
            conternt = "[C]".join(conternt)
            # add the input and target to the data
            text_array.append(conternt)
            target_array.append(target)
        # create a dataframe
        self.data = pd.DataFrame({'text': text_array, 'target': target_array})
        # clean the text
        self.data['text'] = self.data['text'].apply(self.clean_string)
        self.data['target'] = self.data['target'].apply(self.clean_string)
        # tokenize the data with padding to 512 and truncation to 512 (from the end)
        self.data['text'] = self.data['text'].apply(self.tokenize_string)
        self.data['target'] = self.data['target'].apply(self.tokenize_string, target=True)
        # get text ids and attention masks
        self.data['text_input_ids'] = self.data['text'].apply(lambda x: x['input_ids'])
        self.data['text_attention_mask'] = self.data['text'].apply(lambda x: x['attention_mask'])
        self.data = self.data.drop(columns=['text'])
        # get traget ids
        self.data['target'] = self.data['target'].apply(lambda x: x['input_ids'])
        return self.data

    # process a single string
    def clean_string(self, string):
        """
        Cleans a string by removing certain spaces and new line characters.

        :param string: The string to clean.
        :type string: str

        :returns: The cleaned string.
        :rtype: str
        """
        string = string.replace("\n", " ")
        # remove spaces at the beginning and end of the string
        string = string.strip()
        # remove spaces before and after punctuation
        string = string.replace(" .", ".")
        string = string.replace(" ,", ",")
        string = string.replace(" !", "!")
        string = string.replace(" ?", "?")
        string = string.replace(" '", "'")
        string = string.replace(" ’", "’")
        return string

    # tokenize a single string
    def tokenize_string(self, string, target=False):
        """
        Tokenizes a string.

        :param string: The string to tokenize.
        :type string: str

        :param target: Indicates whether the string is a target.
        :type target: bool, optional

        :returns: Tokenized string.
        :rtype: Dict[str, Tensor]
        """
        tokens_obj = self.tokenizer(string, padding='max_length', truncation=True, max_length=512, return_tensors="pt")
        if target:
            tokens_obj['input_ids'][tokens_obj['input_ids'] == self.tokenizer.pad_token_id] = -100
        return tokens_obj

    # decode a single string
    def decode_tokens(self, tokens):
        """
        Decodes a sequence of tokens.

        :param tokens: The tokens to decode.
        :type tokens: Tensor

        :returns: The decoded string.
        :rtype: str
        """
        return self.tokenizer.decode(tokens, skip_special_tokens=True)

    # clean then tokenize a single string
    def process_string(self, string):
        """
        Cleans and tokenizes a conversational string.

        :param string: The conversational string to process.
        :type string: str

        :returns: Processed string in token vector form.
        :rtype: Dict[str, Tensor]
        """
        string = self.clean_string(string)
        tokens_vector = self.tokenize_string(string)
        return tokens_vector


"""
This module contains the configuration class for BERT.
"""

# Bert configuration
class BERTConfig(object):
    """
    Configuration class for BERT.

    This class holds the configuration parameters for the BERT model.

    :param vocab_size: The size of the vocabulary, defaults to 30522.
    :type vocab_size: int
    :param hidden_size: The hidden size of the BERT model, defaults to 768.
    :type hidden_size: int
    :param encoder_layers: The number of encoder layers in the BERT model, defaults to 12.
    :type encoder_layers: int
    :param heads: The number of attention heads in the BERT model, defaults to 12.
    :type heads: int
    :param ff_size: The size of the feed-forward layer in the BERT model, defaults to 3072.
    :type ff_size: int
    :param token_types: The number of token types in the BERT model, defaults to 2.
    :type token_types: int
    :param max_seq: The maximum sequence length in the BERT model, defaults to 512.
    :type max_seq: int
    :param padding_idx: The padding index used in the BERT model, defaults to 0.
    :type padding_idx: int
    :param layer_norm_eps: The epsilon value for layer normalization in the BERT model, defaults to 1e-12.
    :type layer_norm_eps: float
    :param dropout: The dropout rate in the BERT model, defaults to 0.1.
    :type dropout: float
    """
    def __init__(self, vocab_size=30522, hidden_size=768, encoder_layers=12, heads=12, ff_size=3072, token_types=2, max_seq=512, padding_idx=0, layer_norm_eps=1e-12, dropout=0.1):
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.encoder_layers = encoder_layers
        self.heads = heads
        self.ff_size = ff_size
        self.token_types = token_types
        self.max_seq = max_seq
        self.padding_idx = padding_idx
        self.layer_norm_eps = layer_norm_eps
        self.dropout = dropout


from torch import nn

class LinearClassifier(nn.Module):
    '''An interface for a simple linear classifier intended to be used with the BERT model'''
    def __init__(self, input_embeddings_size, output_size):
        super(LinearClassifier, self).__init__()
        # input_embeddings_size as this is gets its input from the output of the BERT model
        """
        Constructs a LinearClassifier instance with specific layer sizes.

        :param input_embeddings_size: The size of each input embeddings from BERT.
        :type input_embeddings_size: int
 
        :param output_size: The size of the output from the model.
        :type output_size: int
 
        :returns: None
        """
        self.linear = nn.Linear(input_embeddings_size, output_size)
        self.softmax = nn.LogSoftmax(dim=1)
        
    def forward(self, input):
        """
        Defines the computation performed by the model.

        :param input: The provided input from the BERT model.
        :type input: Tensor

        :returns: Output after the forward pass (classes probabilities).
        :rtype: Tensor
        """
        output = self.linear(input)
        output = self.softmax(output)
        return output


try:
    from gtts import gTTS
    import os
    from tqdm import tqdm
    from pydub import AudioSegment
    import librosa
    import random
    import soundfile as sf
    import gdown
    import shutil
except:
    print("You need to install pip install botiverse[voice] to use the Vocalize transformer.")

class Vocalize():
    '''
    An interface for transforming words into audio files via Google's Text-to-Speech API and adding noise to them.
    '''
    def __init__(self, words):
        '''
        Initialize the Vocalize transformer by setting the input words and making the dataset.
        :param: words: A list of words to be transformed into audio files
        '''
        self.words = words
        self.make_dataset()
        
    def make_dataset(self):
        '''
        Make a dataset of audio files for the given words by using Google's Text-to-Speech API to pronounce the word
        in australian, british, american, indian, and south african accents.
        '''
        # if there is a folder called dataset, delete it
        if os.path.exists('dataset'): shutil.rmtree('dataset')
        # make a folder for each word in folder 'dataset
        for word in self.words:
            if not os.path.exists('dataset/' + word):
                os.makedirs('dataset/' + word)

        # Make audio for each word
        print("Making audio files for each word...")
        for word in tqdm(self.words):
            tlds = ["com.au", "co.uk", "us", "co.in", "co.za"]
            for i, tld in enumerate(tlds):
                tts = gTTS(text=word, lang="en", tld=tld, slow=False)                       # Sample rate of 24K
                tts.save(f"dataset/{word}/{i}.mp3")
                sound = AudioSegment.from_mp3(f"dataset/{word}/{i}.mp3")
                sound.export(f"dataset/{word}/{i}.wav", format="wav")
                os.remove(f"dataset/{word}/{i}.mp3")

    @staticmethod
    def corrupt_dataset(words=None, sample_rate=16000, traffic=False, force_download=False):
        '''
        Given a folder 'dataset' with folders each containing audio files, this randomly adds noise to each audio file and saves it
        by applying specific noise introduction logic. If noise is not found locally, it is downloaded from Google Drive.
        :param: words: A list of words to be transformed into audio files (i.e., the folder names)
        :param: sample_rate: The sample rate of the audio files
        '''
        # if words is None then assume they are the folder names in dataset
        if words is None: words = os.listdir('dataset')
        
        curr_dir = os.path.dirname(os.path.abspath(__file__))
        # does not exist or is empty
        if not os.path.exists(f"{curr_dir}/noises") or force_download:
            print("Noises not found. Downloading the noise sounds to be used for augmentation...")
            # if not, download the WaveGlow folder
            f_id = '13sOukAKPjoW1K0Ic-8t49P_nkO1dj1oc' 
            gdown.download(f'https://drive.google.com/uc?export=download&confirm=pbef&id={f_id}', curr_dir + '/noises.zip', quiet=False)
            # extract the folder
            shutil.unpack_archive(curr_dir + '/noises.zip', curr_dir)
            print("Done.")   
            # remove the zip file
            os.remove(curr_dir + '/noises.zip') 
        
        print("Corrupting the dataset...")
        for word in tqdm(words):
            # for each file in the folder
            for file in os.listdir(f"dataset/{word}"):
                noise_added = False
                waveform, sample_rate = librosa.load(f"dataset/{word}/{file}", sr=sample_rate)

                #waveform = librosa.effects.pitch_shift(y=waveform, sr=sample_rate, n_steps=-2)

                # with probability 100% add room noise, 40% cafe noise, and 20% traffic noise
                noise_prob = random.random()
                noise_waveform, sr = librosa.load(os.path.join(curr_dir, "./noises/room.wav"), sr=None)
                noise_type = "room"
                
                if noise_prob < 0.4:
                    noise_waveform, sr = librosa.load(os.path.join(curr_dir, "./noises/cafe.wav"), sr=None)
                    noise_type = "cafe"
                    noise_added = True
                elif noise_prob < 0.6 and not noise_added and traffic:
                    noise_waveform, sr = librosa.load(os.path.join(curr_dir, "./noises/traffic.wav"), sr=None)
                    noise_type = "traffic"

                # Resample noise waveform to match the sample rate of the target waveform
                noise_waveform = librosa.resample(y=noise_waveform, orig_sr=sr, target_sr=sample_rate)

                # Trim the waveforms to match the desired duration
                max_offset = len(noise_waveform) - len(waveform)
                offset = random.randint(0, max_offset)
                noise_waveform = noise_waveform[offset:offset+len(waveform)]

                # Add noise to the waveform
                waveform = waveform + noise_waveform

                # Save the modified waveform with noise
                output_folder = f"dataset/{word}"
                # same name but append the noise type
                output_path = f"{output_folder}/{file.split('.')[0]}_{noise_type}.wav"
                sf.write(output_path, waveform, sample_rate, 'PCM_24')
                
                # remove the original file 
                os.remove(f"dataset/{word}/{file}") 




from botiverse.models.FastSpeech1.FastSpeech import TTS
from botiverse.models.NN.NN import NeuralNet
from botiverse.models.SVM.SVM import SVM
from botiverse.models.LSTM import LSTM

from botiverse.models.BERT.BERT import Bert
from botiverse.models.BERT.config import BERTConfig
from botiverse.models.BERT.utils import LoadPretrainedWeights

from botiverse.models.TRIPPY.TRIPPY import TRIPPY
from botiverse.models.TRIPPY.TRIPPY_DST import TRIPPYDST
from botiverse.models.TRIPPY.config import TRIPPYConfig

from botiverse.models.LSTM.LSTM import LSTMClassifier


"""
This module contains the base code and interface of deep Taskbot .
"""

from botiverse.bots.TaskBot.utils import RandomDP, PriorityDP, TemplateBasedNLG
from botiverse.models.TRIPPY.config import TRIPPYConfig
from botiverse.models.TRIPPY.TRIPPY_DST import TRIPPYDST
from botiverse.models.BERT.config import BERTConfig

import random

class TaskBot:
  """
  Instantiate a Deep Task Oriented Dialogue System chat bot.
  It aims to assist the user in completing certain tasks in specific domains.
  The chat bot can use a Deep learning approach for training and inference.

  :param domains: List of domain names.
  :type domains: list[str]

  :param slot_list: List of slot names.
  :type slot_list: list[str]

  :param start: List of initial system utterances and corresponding system acts.
  :type start: list[dict]

  :param templates: The predefined templates for generating responses.
  :type templates: list[dict]

  :param label_maps: Dictionary of the variants of the slot-values that are mapped to the canonical slot-values.
  :type label_maps: dict[str, list]

  :param policy: The dialogue policy to be used ('Random' or 'Priority').
  :type policy: str

  :param non_referable_slots: List of non-referable slots, defaults to an empty list.
  :type non_referable_slots: list[str]

  :param non_referable_pairs: List of non-referable slot-value pairs, defaults to an empty list.
  :type non_referable_pairs: list[tuple]

  :param from_scratch: Indicates whether to use BERT model implemented from scratch in the library, defaults to False.
  :type from_scratch: bool

  :param BERT_config: The configuration of the BERT model, defaults to BERTConfig().
  :type BERT_config: BERTConfig

  :param TRIPPY_config: The configuration of the TRIPPY model, defaults to TRIPPYConfig().
  :type TRIPPY_config: TRIPPYConfig

  :param verbose: Indicates whether to print the chatbot's state after each inference, defaults to False.
  :type verbose: bool
  """
  def __init__(self, domains=[], slot_list=[], start=[], templates=[], label_maps={}, policy='Priority', non_referable_slots=[], non_referable_pairs=[], from_scratch=True, BERT_config=BERTConfig(), TRIPPY_config=TRIPPYConfig(), verbose=False):
    self.domains = domains
    self.policy = policy
    self.start = start
    self.is_start = True
    self.dst = TRIPPYDST(domains, slot_list, label_maps, non_referable_slots, non_referable_pairs, from_scratch, BERT_config, TRIPPY_config)
    self.dpo = RandomDP() if policy == 'Random' else PriorityDP() if policy == 'Priority' else None
    self.nlg = TemplateBasedNLG(templates)
    self.sys_utter = ''
    self.inform_mem = {}
    self.verbose = verbose

  def save_model(self, model_path):
    """
    Save the trained DST model to the given path.

    :param model_path: Path to save the trained DST model.
    :type model_path: str
    """
    self.dst.save_model(model_path)


  def read_data(self, train_path, dev_path=None, test_path=None):
    """
    Read the training, development and testing data and store them in the chatbot.

    :param train_path: Path to the training data file.
    :type train_path: str

    :param dev_path: Path to the development data file, defaults to None.
    :type dev_path: str

    :param test_path: Path to the testing data file, defaults to None.
    :type test_path: str
    """

    self.train_path = train_path
    self.dev_path = dev_path
    self.test_path = test_path

  def train(self):
    """
    Train the chatbot model with the given training data.

    """
    self.dst.train(self.train_path, self.dev_path, self.test_path)

  def load_dst_model(self, model_path=None, pretrained='sim-R', test_path=None):
    """
    Load a trained DST model from the given path.

    :param model_path: Path to the trained DST model.
    :type model_path: str

    :param pretrained: The pretrained model to be used defaults to 'sim-R'.
    :type pretrained: str

    :param test_path: Path to the testing data file, if applicable, defaults to None.
    :type test_path: str
    """
    self.dst.load_model(model_path, pretrained, test_path)

  def infer(self, user_utter):
    """
    Infer a suitable response to the user's utterance.

    :param user_utter: The user's input utterance.
    :type user_utter: str

    :return: The chatbot's response.
    :rtype: str
    """

    response = None

    if self.is_start and len(self.start) > 0:
      temp = self.start[random.randint(0, len(self.start)-1)]
      response, inform_mem = temp['utterance'], temp['system_act']
      self.sys_utter = response
      self.inform_mem = inform_mem
    else:  
      state = self.dst.update_state(self.sys_utter, user_utter, self.inform_mem)
      action = self.dpo.get_action(state, self.nlg.get_templates_slots())
      if action is not None:
        response, inform_mem = self.nlg.generate(action)
        self.sys_utter = response
        self.inform_mem = inform_mem

    self.is_start = False

    if self.verbose:
      print(f'State: {self.get_dialogue_state()}')

    return response
  
  def infer_with_state(self, user_utter, line_breaker='<br/>'):
    """
    Infer a suitable response to the user's utterance and return the dialogue state.

    :param user_utter: The user's input utterance.
    :type user_utter: str

    :return: The chatbot's response and the dialogue state concatenated into a single string.
    :rtype: str
    """

    response = self.infer(user_utter)
    if response is None:
      response = ''
    state = self.get_dialogue_state()
    return response + line_breaker + str(state)


  def suggest(self, template):
    """
    Set the system utterance and system act to suggest a specific response.

    :param template: The template containing the suggested system utterance and system act.
    :type template: dict
    """
    self.sys_utter = template['utterance']
    self.inform_mem = template['system_act']

  def get_dialogue_state(self):
    """
    Get the dialogue state.

    :return: The dialogue state.
    :rtype: dict
    """
    return self.dst.get_dialogue_state()

  def delete_slots(self, domain=None, slot=None):
    """
    Delete slots from the dialogue state.

    Note that:
    if domain!=None will delete all slots in that domain.
    if slot!=None will delete that slot.
    if both are None will delete all slots in all domains.

    :param domain: The domain from which to delete slots, defaults to None.
    :type domain: str

    :param slot: The slot to delete, defaults to None.
    :type slot: str
    """
    self.dst.delete_slots(domain, slot)

  def reset(self):
    """
    Reset the chatbot's state.
    """
    self.dst.reset()
    self.sys_utter = ''
    self.inform_mem = {}
    self.domain = self.domains[0]
    self.is_start = True

  def __str__(self):
    """
    Return a string representation of the chatbot.

    :return: A string representation of the chatbot.
    :rtype: str
    """
    string = ''
    string = string + '\ndomains: ' + str(self.domains)
    string = string + '\npolicy: ' + str(self.policy)
    string = string + '\nstart: ' + str(self.start)
    string = string + '\nis_start: ' + str(self.is_start)
    string = string + '\n\ndst: ' + str(self.dst)
    string = string + '\n\ndpo: ' + str(self.dpo)
    string = string + '\n\nnlg: ' + str(self.nlg)
    string = string + '\n\nsys_utter: ' + str(self.sys_utter)
    string = string + '\ninform_mem: ' + str(self.inform_mem)
    return string


from botiverse.bots.WhizBot.WhizBot_GRU import WhizBot_GRU
from botiverse.bots.WhizBot.WhizBot_BERT import WhizBot_BERT

class WhizBot:
    '''
    A class that provides an interface for the WhizBot-BERT and WhizBot-GRU models.
    '''
    def __init__(self, repr='BERT'):
        """
        Initializes WhizBot and sets its representation type.
        :param repr: The representation type of the WhizBot model. Either "BERT" or "GRU".
        :type repr: str
        """
        if repr == 'BERT':
            self.bot = WhizBot_BERT()
        elif repr == 'GRU':
            self.bot = WhizBot_GRU()
        else:
            raise ValueError('Invalid representation type for WhizBot. Please choose either "BERT" or "GRU".')
    
    
    
    def read_data(self, file_path):
        """
        Reads and pre-processes the data, sets up the model based on the data and prepares the train-validation split.

        :param file_path: The path to the file that contains the dataset.
        :type file_path: str

        :returns: None
        """
        self.bot.read_data(file_path)

    def train(self, epochs=10, batch_size=32):
        """
        Trains the model using the training dataset.

        :param epochs: The number of training epochs.
        :type epochs: int

        :param batch_size: The number of training examples utilized used to make one paramenters updat.
        :type batch_size: int

        :returns: None
        """
        self.bot.train(epochs, batch_size)

    def validation(self, batch_size=32):
        """
        Tests the model performance using the validation dataset and calculates the accuracy.
        
        :param batch_size: The number of training examples utilized used to make one paramenters updat.
        :type batch_size: int

        :returns: None
        """
        self.bot.validation(batch_size)

    def infer(self, string):
        """
        Performs inference using the model.

        :param string: The input string to perform inference on.
        :type string: str

        :returns: A random response from the response list of the predicted label.
        """
        return self.bot.infer(string)

    def save(self, path):
        """
        Saves the model parameters to the given path.

        :param path: The path where the model parameters will be saved.
        :type path: str

        :returns: None
        """
        self.bot.save(path)

    def load(self, path):
        """
        Loads the model parameters from the given path.

        :param path: The path from where the model parameters will be loaded.
        :type path: str

        :returns: None
        """
        self.bot.load(path)


import pandas as pd
import torch
from transformers import AutoModel, AutoTokenizer
from tqdm.auto import tqdm
import re

class WhizBot_BERT_Preprocessor:
    '''An interface that provides the required preprocessing for the WhizBot_BERT bot'''
    def __init__(self, file_path):
        """
        Initializes a WhizBot_BERT_Preprocessor instance with the file path of the dataset and the BERT model parameters.

        :param file_path: Path to the .json file to be read.
        :type file_path: str

        :returns: None
        """
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.file_path = file_path
        self.text_col = "text"
        self.label_col = "label"
        self.data = pd.read_json(self.file_path)
        # make a dictionary of all the tag and responses
        self.responces = {}
        for i in range(len(self.data)):
            self.responces[self.data['tag'][i]] = self.data['responses'][i]
        # remove the responses column from the dataframe
        self.data = self.data.drop('responses', axis=1)
        # stretch the dataframe to have one row per pattern
        self.data = self.data.explode('patterns')
        # rename the columns tag->label and patterns->text
        self.data = self.data.rename(columns={'tag': 'label', 'patterns': 'text'})
        # remove all rows with empty text
        self.data = self.data.dropna()
        self.data = self.data.reset_index(drop=True)
        # load the pretrained model and freeze its parameters (not to be trained)
        self.bert = AutoModel.from_pretrained("bert-base-multilingual-cased").to(self.device)
        for param in self.bert.parameters():
            param.requires_grad = False

    def process(self):
        """
        Applies preprocessing steps to the loaded data.

        :returns: Processed data.
        :rtype: DataFrame
        """
        # text processing
        self.data['text'] = self.data['text'].apply(self.clean_string)
        # tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained("bert-base-multilingual-cased")
        # tokenize the data
        tqdm.pandas(desc='Tokenizing strings')
        self.data['text'] = self.data['text'].progress_apply(self.tokenize_string)
        # embed the tokens
        tqdm.pandas(desc='Embedding tokens')
        self.data['text'] = self.data['text'].progress_apply(self.embed_tokens)
        # make a dictionary of all labels
        self.label_dict = {}
        for i in range(len(self.data)):
            if self.data['label'][i] not in self.label_dict:
                self.label_dict[self.data['label'][i]] = len(self.label_dict)
        # now replace all labels with their corresponding index
        self.data['label'] = self.data['label'].apply(lambda x: self.label_dict[x])
        # now convert to tensors
        self.data['label'] = self.data['label'].apply(lambda x: torch.tensor([x]))
        return self.data

    # process a single string
    def clean_string(self, string):
        """
        Cleans the given text string by removing the emojies.

        :param string: The string to process.
        :type string: str

        :returns: The processed string.
        :rtype: str
        """
        # remove the emoji
        emoji_pattern = re.compile("["
            u"\U0001F600-\U0001F64F"
            u"\U0001F300-\U0001F5FF"
            u"\U0001F680-\U0001F6FF"
            u"\U0001F1E0-\U0001F1FF"
            u"\U00002500-\U00002BEF"
            "]+", flags=re.UNICODE)
        string = emoji_pattern.sub(r'', string)
        return string

    # tokenize a single string
    def tokenize_string(self, string):
        """
        Tokenizes a given text string using the BERT tokenizer.

        :param string: The string to tokenize.
        :type string: str

        :returns: A dictionary containing the tokenized version of the input text strin i.e., the ids and attention_masks.
        :rtype: dict
        """
        tokens_obj = self.tokenizer(string, padding=True, truncation=True, max_length=256, return_tensors="pt")
        return tokens_obj

    # as the model is frozen, we can make the training way faster by precomputing the embeddings, with that it will only be computed once
    def embed_tokens(self, tokens_obj):
        """
        Precomputes embeddings for the tokenized text string.

        :param tokens_obj: A dictionary containing the tokenized version of a text string.
        :type tokens_obj: dict

        :returns: Tensor representing the embeddings of the tokenized text string.
        :rtype: Tensor
        """
        # average of all the output embeddings (to capture the whole sentence meaning)
        embeddings = self.bert(input_ids=tokens_obj['input_ids'].to(self.device), attention_mask=tokens_obj['attention_mask'].to(self.device),return_dict=False)[0].mean(dim=1).detach().cpu()
        return embeddings

    # the whole preprocessing pipeline
    def process_string(self, string):
        """
        Applies the whole preprocessing pipeline to a given text string.

        :param string: The string to process
        :type string: str

        :returns: Tensor representing the embeddings of the processed and tokenized text string.
        :rtype: Tensor
        """
        string = self.clean_string(string)
        tokens_obj = self.tokenize_string(string)
        embeddings = self.embed_tokens(tokens_obj)
        return embeddings


"""
This Module has the inference functions for TRIPPY.
"""


import torch

from botiverse.models.TRIPPY.data import create_inputs
from botiverse.models.TRIPPY.utils import create_span_output


def infer(model, slot_list, current_state, history, sys_utter, user_utter, inform_mem, device, oper2id, tokenizer, max_len):
  """
  Infer the dialogue state using the TRIPPY model.

  :param model: The TRIPPY model for inference.
  :type model: TRIPPY
  :param slot_list: The list of slots.
  :type slot_list: list
  :param current_state: The current dialogue state.
  :type current_state: dict
  :param history: The dialogue history.
  :type history: list
  :param sys_utter: The system's utterance.
  :type sys_utter: str
  :param user_utter: The user's utterance.
  :type user_utter: str
  :param inform_mem: The inform memory.
  :type inform_mem: dict
  :param device: The device to run the inference on.
  :type device: torch.device
  :param oper2id: The mapping of operations to IDs.
  :type oper2id: dict
  :param tokenizer: The tokenizer to tokenize the input.
  :type tokenizer: transformers.PreTrainedTokenizer
  :param max_len: The maximum length of the input sequence.
  :type max_len: int
  :return: The predicted dialogue state.
  :rtype: dict
  """

  model.eval()

  # turn data to inputs
  input, ids, mask, token_type_ids, tok_input_offsets, input_tokens, padding_len = create_inputs(history,
                                                                                                 user_utter,
                                                                                                 sys_utter,
                                                                                                 tokenizer,
                                                                                                 max_len)


  # print(input, ids, mask, token_type_ids, tok_input_offsets, input_tokens, padding_len)


  with torch.no_grad():
    n_slots = len(slot_list)
    ids = torch.tensor(ids, dtype=torch.long).unsqueeze(0).to(device)
    mask = torch.tensor(mask, dtype=torch.long).unsqueeze(0).to(device)
    token_type_ids = torch.tensor(token_type_ids, dtype=torch.long).unsqueeze(0).to(device)
    inform_aux_features = torch.zeros((1, n_slots)).to(device)
    ds_aux_features = torch.zeros((1, n_slots)).to(device)
    input_tokens = ' '.join(input_tokens)
    padding_len = padding_len

    for slot_idx, slot in enumerate(slot_list):
      if slot in inform_mem:
        inform_aux_features[0, slot_idx] = 1
      if slot in current_state:
        ds_aux_features[0, slot_idx] = 1

    # print(slot_list)
    # print(inform_aux_features)
    # print(inform_mem)


    # get model outputs
    slots_start_logits, slots_end_logits, slots_oper_logits, slots_refer_logits = model(ids=ids,
                                                                                        mask=mask,
                                                                                        token_type_ids=token_type_ids,
                                                                                        inform_aux_features=inform_aux_features,
                                                                                        ds_aux_features=ds_aux_features)


    # update the predicted state of each slot
    pred_state = current_state.copy()
    for slot_idx, slot in enumerate(slot_list):

      # get the predicted operation
      pred_oper = slots_oper_logits[slot_idx][0].argmax(dim=-1).item()
      # print(slot, torch.softmax(slots_oper_logits[slot_idx][0], dim=-1))

      # update the slot based on the operation
      if pred_oper == oper2id['carryover']: # carryover
        continue
      elif pred_oper == oper2id['dontcare']: # dontcare
        pred_state[slot] = 'dontcare'
      elif pred_oper == oper2id['update']: # update
        pred_state[slot] = create_span_output(slots_start_logits[slot_idx][0].cpu().detach().numpy(),
                                              slots_end_logits[slot_idx][0].cpu().detach().numpy(),
                                              padding_len,
                                              input_tokens)
      elif pred_oper == oper2id['refer']: # refer
        refered_slot = slots_refer_logits[slot_idx][0].argmax(dim=-1).item()
        if refered_slot != n_slots and slot_list[refered_slot] in current_state:
          pred_state[slot] = current_state[slot_list[refered_slot]]
      elif pred_oper == oper2id['yes']: # yes
        pred_state[slot] = 'yes'
      elif pred_oper == oper2id['no']: # no
        pred_state[slot] = 'no'
      elif pred_oper == oper2id['inform']: # inform
        if slot in inform_mem:
          pred_state[slot] = '§§' + inform_mem[slot][0]


  return pred_state


from gensim.utils import tokenize
import numpy as np
from nltk.stem.porter import PorterStemmer
stemmer = PorterStemmer()

from botiverse.preprocessors import GloVe, TF_IDF

class TF_IDF_GLOVE():
    '''
    An interface for transforming sentences into glove vectors by weighting word GloVe vectors by their tf-idf.
    '''
    def __init__(self, force_download=False):
        '''
        Initialize the GloVe and TF-IDF transformer and download the embeddings if needed.
        '''
        self.glove = GloVe(force_download)
        self.glove_dict = self.glove.glove_dict
        self.tf_idf = TF_IDF()
        #self.glove_dict = None
        self.tf = None
        self.idf = None
        self.all_words = None


    def transform_list(self, sentence_list, all_words):
        '''
        Given a list of tokenized sentences, return a table of GloVe vectors (one for each sentence) in the form of a numpy array.
        This also initializes the tf and idf tables of the class for use in the transform() method.
        :param: sentence_list: A list of tokenized sentences
        :param: all_words: A list of all the words in the corpus
        '''
        self.all_words = all_words
        # just to set tf and idf
        self.tf_idf.transform_list(sentence_list, all_words)
        self.tf, self.idf = self.tf_idf.tf, self.tf_idf.idf
        # make a numpy array of the sentence vectors
        X = np.zeros((len(sentence_list), 50))
        for i, sentence in enumerate(sentence_list):
                X[i] = self.transform(sentence)
        return X

    def transform(self, sentence):
        '''
        Given a sentence, return its GloVe vector as a numpy array by weighting the GloVe vectors of the words in the sentence by their idf then averaging.
        :param: sentence: A string of words
        :return: A numpy array of the GloVe vector
        '''
        tokens = list(tokenize(sentence, to_lower=True))
        tokens_s = [stemmer.stem(word.lower()) for word in tokens if word not in ['?', '!', '.', ',']]

        # get the idf of each word in the sentence
        weights = np.zeros((len(tokens)))
        for i, word in enumerate(tokens_s):
            # if word in self.all_words:
            if word in self.all_words and tokens[i] in self.glove_dict:
                weights[i] = self.idf[self.all_words.index(word)]
        
        # normalize the weights
        weights = weights / np.sum(weights)
        
        # get the weighted average of the glove vectors
        avg_vector = np.zeros(50)
        for i, word in enumerate(tokens):
            if word in self.glove_dict:
                avg_vector += weights[i] * self.glove_dict[word]
        
        avg_vector = avg_vector[np.newaxis, :]
        return avg_vector



import torch
import torch.nn as nn

class BasicGRU(nn.Module):
    '''An interface for the basic GRU unit'''
    def __init__(self, input_size, dropout_p=0.1):
        """
        Constructs a BasicGRU instance with specific layer sizes and dropout probability.
        
        :param input_size: The size of the input to the model.
        :type input_size: int

        :param dropout_p: A regularization parameter.
        :type dropout_p: float

        :returns: None
        """
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        super(BasicGRU, self).__init__()
        self.hidden_size = input_size
        # update gate
        self.W_z = nn.Linear(input_size, input_size)
        # reset gate
        self.W_r = nn.Linear(input_size, input_size)
        # new memory gate
        self.W_h = nn.Linear(input_size, input_size)
        # activation functions
        self.sigmoid = nn.Sigmoid()
        self.tanh = nn.Tanh()
        # dropout (for regularization)
        self.dropout = nn.Dropout(dropout_p)

    def forward(self, input, hidden):
        """
        Defines the computation performed by the model.

        :param input: The provided sequance input.
        :type input: Tensor
        
        :param hidden: The provided hidden state.
        :type hidden: Tensor

        :returns: New hidden state.
        :rtype: Tensor
        """
        combined = input + hidden
        # update activation vector calculation
        z = self.W_z(combined)
        z = self.sigmoid(z)
        # reset activation vector calculation
        r = self.W_r(combined)
        r = self.sigmoid(r)
        # new memory vector calculation
        h = self.W_h(input + r * hidden)
        h = self.tanh(h)
        # set the memory with the weighted sum of old and new memory
        hidden = (1 - z) * hidden + z * h
        # apply dropout
        hidden = self.dropout(hidden)
        return hidden

    # initialize the hidden state with zeros (initial memory)
    def initHidden(self, batch_size):
        """
        Creates a tensor of zeros for the hidden state initialization.

        :param batch_size: The size of the batch for which the hidden state is to be initialized.
        :type batch_size: int

        :returns: Tensor of zeros of the shape (batch_size, 1, hidden_size).
        :rtype: Tensor
        """
        return torch.zeros(batch_size, 1, self.hidden_size).to(self.device)

class GRUTextClassifier(nn.Module):
    ''' An interface for the GRU text classifier which uses a basic GRU unit with a linear output layer and an input embedding layer'''
    def __init__(self, vocabulary, embedding_size, output_size, dropout_p=0.1):
        """
        Constructs a GRUTextClassifier instance with specific hyperparameters.
 

        :param vocabulary: The size of vocabulary used in the Embedding layer.
        :type vocabulary: int
 
        :param embedding_size: The size of each embedding vector.
        :type embedding_size: int
 
        :param output_size: The size of the output from the model (number of classes).
        :type output_size: int

        :param dropout_p: A regularization parameter.
        :type dropout_p: float
 
        :returns: None
        """
        super(GRUTextClassifier, self).__init__()
        # the embedding layer
        self.embedding = nn.Embedding(vocabulary, embedding_size)
        # the GRU layer
        self.gru = BasicGRU(embedding_size, dropout_p)
        # the output layer
        self.h2o = nn.Linear(embedding_size, output_size)
        # activation function
        self.softmax = nn.LogSoftmax(dim=1)

    def forward(self, input):
        """
        Defines the computation performed by the model.

        :param input: The model input.
        :type input: Tensor

        :returns: Output after the forward pass (classes probabilities).
        :rtype: Tensor
        """
        batch_size = input.size()[0]
        # get the first hidden state
        hidden = self.gru.initHidden(batch_size)
        # embed the input
        input_temp = self.embedding(input)
        # pass the input through the GRU layer for each token in the sequence
        for i in range(input_temp.size()[1]):
            hidden = self.gru(input_temp[:, i:i+1, :], hidden)
        # pass the last hidden state through the output layer
        output = self.h2o(hidden.squeeze(1))
        # apply softmax
        output = self.softmax(output)
        return output



"""
This Module has the configuration class for TRIPPY.
"""

import tokenizers
import os
import gdown

# Trippy configuration
class TRIPPYConfig(object):
    """
    Configuration class for TRIPPY.

    This class holds the configuration parameters for the TRIPPY model.

    :param max_len: The maximum sequence length, defaults to 128.
    :type max_len: int
    :param train_batch_size: The batch size for training, defaults to 32.
    :type train_batch_size: int
    :param dev_batch_size: The batch size for development evaluation, defaults to 1.
    :type dev_batch_size: int
    :param test_batch_size: The batch size for testing, defaults to 1.
    :type test_batch_size: int
    :param epochs: The number of training epochs, defaults to 15.
    :type epochs: int
    :param hid_dim: The hidden dimension size, defaults to 768.
    :type hid_dim: int
    :param n_oper: The number of operations, defaults to 7.
    :type n_oper: int
    :param dropout: The dropout rate, defaults to 0.3.
    :type dropout: float
    :param vocab_path: The path to the vocabulary file, defaults to 'vocab.txt'.
    :type vocab_path: str
    :param ignore_idx: The index value to ignore, defaults to -100.
    :type ignore_idx: int
    :param oper2id: The mapping of operation names to IDs, defaults to {'carryover' : 0, 'dontcare': 1, 'update':2, 'refer':3, 'yes':4, 'no':5, 'inform':6}.
    :type oper2id: dict[str, int]
    :param weight_decay: The weight decay value, defaults to 0.0.
    :type weight_decay: float
    :param lr: The learning rate, defaults to 1e-4.
    :type lr: float
    :param adam_epsilon: The epsilon value for Adam optimizer, defaults to 1e-6.
    :type adam_epsilon: float
    :param warmup_proportion: The proportion of warmup steps, defaults to 0.1.
    :type warmup_proportion: float
    :param multiwoz: The path to the MultiWOZ dataset, defaults to False.
    :type multiwoz: str
    """
    def __init__(self, 
                 max_len=128, 
                 train_batch_size=32, 
                 dev_batch_size=1, 
                 test_batch_size=1, 
                 epochs=15, 
                 hid_dim=768, 
                 n_oper=7, 
                 dropout=0.3,
                 vocab_path='vocab.txt',
                 ignore_idx=-100,
                 oper2id={'carryover' : 0, 'dontcare': 1, 'update':2, 'refer':3, 'yes':4, 'no':5, 'inform':6},
                 weight_decay=0.0,
                 lr=1e-4,
                 adam_epsilon=1e-6,
                 warmup_proportion=0.1,
                 multiwoz=False):

        self.max_len = max_len
        self.train_batch_size = train_batch_size
        self.dev_batch_size = dev_batch_size
        self.test_batch_size = test_batch_size
        self.epochs = epochs
        self.hid_dim = hid_dim
        self.n_oper = n_oper
        self.dropout = dropout
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        self.vocab_path = os.path.join(cur_dir, vocab_path)
        if not os.path.exists(self.vocab_path):
            print("Downloading Vocab...")
            f_id = '1f2iOTT-QiFbIc1naqGVZWX5wPVo7gUMS' 
            gdown.download(f'https://drive.google.com/uc?export=download&confirm=pbef&id={f_id}', self.vocab_path, quiet=False)
            print("Done.")   
        self.tokenizer = tokenizers.BertWordPieceTokenizer(self.vocab_path, lowercase=True)
        self.ignore_idx = ignore_idx
        self.oper2id = oper2id
        self.weight_decay = weight_decay
        self.lr = lr
        self.adam_epsilon = adam_epsilon
        self.warmup_proportion = warmup_proportion
        self.multiwoz = multiwoz


import numpy as np
import json
from gensim.utils import tokenize
import numpy as np
from botiverse.models import SVM, NeuralNet
from botiverse.preprocessors import GloVe, TF_IDF, TF_IDF_GLOVE, BoW
from nltk.stem.porter import PorterStemmer
stemmer = PorterStemmer()


class BasicBot:
    '''
    An interface for a basic chatbot model suitable for small datasets such as FAQs. Note that the
    underlying model is not sequential (either an NN or an SVM).
    '''
    def __init__(self, machine='nn', repr='tf-idf'):
        """
        Instantiate a basic chat bot model that uses a classic feedforward neural network.
        Data can be then used to train the chatbot model.
        
        :param name: The chatbot's name.
        :type name: string
        :param machine: The machine learning model to use. Either 'nn' or 'svm', else must be a model object that has a fit method.
        :type machine: string
        :param repr: The representation to use. Either 'glove', 'tf-idf', or 'tf-idf-glove' or 'bow', else must be a model object that has a transform method.
        :type repr: string
        """
        self.model = None
        self.machine = machine
        self.repr = repr
        
        if repr == 'glove': 
            self.transformer = GloVe()
        elif repr == 'tf-idf':
            self.transformer = TF_IDF()
        elif repr == 'tf-idf-glove':
            self.transformer = TF_IDF_GLOVE()
        elif repr == 'bow':
            self.transformer = BoW()
        elif type(repr) != str:
            # if machine or transform is not a string, then assume it is a model
            self.transformer = repr
        else:
            raise Exception('Representation must either be one of those the basic chatbot support or a custom one that implement the transform API. Found was ' + repr)
            
        self.tf = None
        self.idf = None
        self.classes = None
        
        

    def setup_data(self):
        """
        Given JSON data, set up the data for training by converting it to a list of sentences and their corresponding classes.
        """  
        all_words = []
        classes = []
        sentence_list = []                             # sentence_table[i] is a tuple (list of words, class)
        y = []
        for intent in self.raw_data:                    #this is a list of dictionaries. each has a tag (class), list of patterns and list of responses.
            tag = intent['tag']
            classes.append(tag)
            for pattern in intent['patterns']:
                if self.repr == 'tf-idf' or self.repr == 'tf-idf-glove' or self.repr == 'bow':
                    all_words += list(tokenize(pattern, to_lower=True))     
                sentence_list.append(pattern)
                y.append(tag)

        # stem and lower each word
        all_words = [stemmer.stem(word.lower()) for word in all_words if word not in ['?', '!', '.', ',']]
        
        # remove duplicates and sort alphabetically
        all_words = sorted(set(all_words))
        classes = sorted(set(classes))
        
        self.all_words = all_words
        self.classes = classes
        
        X = self.transformer.transform_list(sentence_list, all_words=all_words)

        # convert each class to its index
        for i, tag in enumerate(y):
            y[i] = classes.index(tag)
        y = np.array(y)
        return X, y
    
    def read_data(self, path):
        """
        Read the data from a JSON file for the chatbot to train on later.
        :param data: A stringfied JSON object containing the training data 
        :type number: string
        """
        with open(path, 'r') as f:
            self.raw_data = json.load(f) 

        self.X, self.y = self.setup_data()
        
    def train(self, max_epochs=None, early_stop=False, **kwargs):
        """
        Train the chatbot model with the given JSON data.
        
        :param early_stop: Whether to use early stopping or not
        :type early_stop: bool
        :param provided_model: A model to use instead of the default one
        :type provided_model: Object
        :param provided_params: A dictionary of parameters to use instead of the default ones
        :type provided_params: dict
    
        :return: None
        :rtype: NoneType
        """
        X, y = self.X, self.y
        if self.machine == 'nn':
            self.model = NeuralNet(structure=[X.shape[1], 12, len(self.classes)], activation='sigmoid')
            max_epochs = max_epochs if max_epochs is not None else 30 * len(self.classes)
            if early_stop:
                self.model.fit(X, y, batch_size=1, epochs=max_epochs, λ = 0.02, eval_train=True, val_split=0.2, patience=100)
                self.model.fit(X, y, batch_size=1, epochs=max_epochs, λ = 0.02, eval_train=True, val_split=0.0)
            else:
                self.model.fit(X, y, batch_size=1, epochs=max_epochs, λ = 0.02, eval_train=True, val_split=0.0)
                
        elif self.machine == 'svm':
            self.model = SVM(kernel='linear', C=700)
            self.model.fit(X, y, eval_train=True)
        elif type(self.machine) != str:
                self.machine.fit(X, y, **kwargs)
        else:
            raise Exception('Machine must either be one of those the basic chatbot support or a custom one that implement the fit API. Found was ' + self.machine)


    def save(self, path):
        '''
        Save the model to a file.
        :param path: The path to the file
        '''
        if self.machine == 'svm':
            print("Could Not Save: SVM model is for experimentation only and does not allow saving yet.")
        else:
            self.model.save(path+'.bot')
    
    def load(self, load_path, data_path):
        '''
        Load the model from a file.
        :param path: The path to the file
        '''
        if self.machine != 'nn':
            print("Could Not Load: SVM or custom model is for experimentation only and does not allow loading yet.")
        else:
            self.model = NeuralNet.load(load_path + '.bot')
            # following can be optimized.
            with open(data_path, 'r') as f:
                self.raw_data = json.load(f)
                self.classes = sorted(set([intent['tag'] for intent in self.raw_data]))
                
                # compute all words again 
                all_words = []
                for intent in self.raw_data:
                    for pattern in intent['patterns']:
                        all_words += list(tokenize(pattern, to_lower=True))
                all_words = [stemmer.stem(word.lower()) for word in all_words if word not in ['?', '!', '.', ',']]
                self.all_words = sorted(set(all_words))
                # set the transformer's all_words if needed
                if hasattr(self.transformer, 'all_words'):  self.transformer.all_words = self.all_words
                    
                
    def infer(self, prompt, confidence=None, test=False):
        """
        Infer a suitable response to the given prompt.
        
        :param promp: The user's prompt
        :type number: string
    
        :return: The chatbot's response
        :rtype: string
        """
        if confidence is None: confidence = 2/len(self.classes)
        vector = self.transformer.transform(prompt) 
        # predict the class of the prompt
        tag_idx, tag_prob = self.model.predict(vector)
        tag_idx, tag_prob = tag_idx[0], tag_prob[0]
        tag = self.classes[tag_idx]
        if tag_prob < confidence: return "Could you rephrase that?"
        for intent in self.raw_data:
            if tag == intent["tag"]:
                if test: return tag
                return np.random.choice(intent['responses'])